"""Provides Utilities to calculate IR metrics given query-passage qrels and scoress."""

from collections import defaultdict
from typing import Dict, List, Optional, Set, Tuple, Union

import torch
from transformers import EvalPrediction, TrainingArguments

from .relevance_evaluator import RelevanceEvaluatorPlus


class IRMetrics:
    def __init__(
        self,
        k_values: Optional[Union[int, List[int]]] = None,
        measures: Optional[Set[str]] = None,
        relevance_threshold: int = 1,
        include_in_batch_negs: bool = False,
    ) -> None:
        """Calculates IR metrics given qrels and scores.

        * you can either use this class by calling :meth:`compute` which just returns the average metrics and does not change the instance internal state.
        * You can call :meth:`add_batch` multiple times and then call :meth:`aggregate_results` to get the average metrics across all examples seen so far.
        * You can also pass an instance of this class to ``transformers.Trainer`` as the ``compute_metrics`` argument.
          But, you have to set ``batch_eval_metrics=True`` to be compatible with this class.

        This is how to use it::

            metric_calculator = IRMetrics(k_values=[10, 100])
            args = TrainingArguments(..., batch_eval_metrics=True, label_names=["YOUR_LABEL_NAME"], ...)
            trainer = Trainer(args=args, ..., compute_metrics=compute_metrics, ...)
            ...

        Args:
            k_values (Union[int, List[int]]): a single or a list of cutoff values to calculate IR metrics for.
                E.g., nDCG@K or recall@K. If provided, use a predefined set of IR metrics with these cutoff values.
            measures (Optional[Set[str]]): a set of measures used with ``pytrec_eval``. You should either specify ``measures`` directly or specify ``k_values``
                and we instantiate a default set of IR metrics with the given ``k_values``.
            relevance_threshold (int): minimum groundtruth relevancy level (inclusive) for a document to be considered relevant.
                Basically documents with relevancy levels smaller than this value will be excluded from the groundtruth.
                See see ``pytrec_eval.RelevanceEvaluator.__init__()`` for details.
            include_in_batch_negs (bool): If true, include the in-batch negatives in metric calculation if available.
                It is only used if an instance of the class is passed to ``transformers.Trainer`` for `compute_metrics` argument.
        """
        self.include_in_batch_negs = include_in_batch_negs
        if self.include_in_batch_negs:
            raise NotImplementedError

        if (k_values is None) == (measures is None):
            msg = f"You should specify one and only one of the 'k_values' or 'measures' arguments. Got measures: '{measures}' and k_values: '{k_values}'"
            raise ValueError(msg)

        self.relevance_threshold = relevance_threshold
        if measures is not None:
            self.metric_ids = measures
        else:
            if isinstance(k_values, int):
                k_values = [k_values]
            if not isinstance(k_values, list):
                raise TypeError
            # Identifier for IR Metrics to calculate. These are later used with pytrec_eval library
            map_string = "map_cut." + ",".join([str(k) for k in k_values])
            ndcg_string = "ndcg_cut." + ",".join([str(k) for k in k_values])
            recall_string = "recall." + ",".join([str(k) for k in k_values])
            precision_string = "P." + ",".join([str(k) for k in k_values])
            mrr_string = "mrr_cut." + ",".join([str(k) for k in k_values])
            self.metric_ids = {
                map_string,
                ndcg_string,
                recall_string,
                precision_string,
                mrr_string,
            }

        # keep track of the sum of metrics for all examples seen so far.
        self.metric_sum = defaultdict(float)
        # number of examples seen so far
        self.num_examples = 0

    def reset_state(self) -> None:
        """Clear the metrics collected so far.

        Useful if you are using the same instance for multiple eval loops.
        """
        del self.metric_sum
        self.metric_sum = defaultdict(float)
        self.num_examples = 0

    def _compute_metrics_sum(
        self,
        scores: Dict[str, Dict[str, Union[int, float]]],
        qrels: Dict[str, Dict[str, int]],
    ) -> Tuple[Dict[str, float], int]:
        """Calculates IR metrics for the given qrels and scores and returns the sum of the metrics
        across all queries.

        This function does not modify the internal state of the class.

        Args:
            scores (Dict[str, Dict[str, Union[int, float]]]): a mapping where ``scores[qid][docid]`` is the predicted similarity score between query ``qid`` and document ``docid``
            qrels (Dict[str, Dict[str, int]]): The ground truth relevance scores. ``qrels[qid][docid]`` is the groundtruth similarity score for query ``qid`` and document ``docid``

        Returns:
            A tuple. First item is the sum of calculated IR metrics for the given examples.
            The second item is the number of examples passed to this function in the current call.
        """
        current_metric_sum = defaultdict(float)
        current_num_examples = 0

        evaluator = RelevanceEvaluatorPlus(
            qrels, self.metric_ids, self.relevance_threshold
        )
        all_evals = evaluator.evaluate(scores)

        # aggregate the metrics for queries in this function call
        for query_eval in all_evals.values():
            for metric_name, metric_value in query_eval.items():
                current_metric_sum[metric_name] += metric_value
            current_num_examples += 1

        return current_metric_sum, current_num_examples

    def compute(
        self,
        scores: Dict[str, Dict[str, Union[int, float]]],
        qrels: Dict[str, Dict[str, int]],
    ) -> Dict[str, float]:
        """Calculates the average IR metrics for the given scores and qrels. Does not modify the
        internal state of the class.

        Args:
            scores (Dict[str, Dict[str, Union[int, float]]]): a mapping where ``scores[qid][docid]`` is the predicted similarity score between query ``qid`` and document ``docid``
            qrels (Dict[str, Dict[str, int]]): The ground truth relevance scores. ``qrels[qid][docid]`` is the groundtruth similarity score for query ``qid`` and document ``docid``

        Returns:
            A mapping from metric name to the average value of the metric for the given examples.
        """
        metric_sums, num_samples = self._compute_metrics_sum(scores=scores, qrels=qrels)

        current_avg = dict()
        for k, v in metric_sums.items():
            current_avg[k] = v / num_samples
        return current_avg

    def add_batch(
        self,
        scores: Dict[str, Dict[str, Union[int, float]]],
        qrels: Dict[str, Dict[str, int]],
    ) -> Dict[str, float]:
        """Calculates the average metrics for the given scores and qrels.

        It also updates the sum of the metric values and total number of examples in the instance's internal state.
        However, in each individual call, it only returns the average metrics for examples in the given call and not across all examples seen so far.

        See :meth:`compute` for description of arguments.

        Returns:
            The average IR metrics for examples in this batch.
        """
        _sum, num_samples = self._compute_metrics_sum(scores=scores, qrels=qrels)
        # Calculate the current average metrics and add them to the internal state of the metrics as well
        current_avg = dict()
        self.num_examples += num_samples
        for k, v in _sum.items():
            self.metric_sum[k] += v
            current_avg[k] = v / num_samples

        return current_avg

    def aggregate_results(self) -> Dict[str, float]:
        """Calculates the average IR metrics across all examples seen so far."""
        avg = dict()
        for _name, _sum in self.metric_sum.items():
            avg[_name] = _sum / self.num_examples
        return avg

    def __call__(
        self, eval_pred: EvalPrediction, compute_result: Optional[bool] = None
    ) -> Dict[str, float]:
        """Calculates the IR metrics given the ``eval_pred`` results from ``transformers.Trainer``.

        This is intended to be used by ``transformers.Trainer`` to calculate the metrics for each batch in evaluation loop.

        * The input logits (i.e., similarity scores) are expected to be of shape `[NUM_QUERIES, NUM_QUERIES * NUM_DOCS_PER_QUERY]`
        * The columns are ordered according to queries. I.e., all passages related to first query show up first, then passages related to second query, and so on.
        * In a distributed environment where ``query_embs`` and ``passage_embs`` are gathered from all processes before calculating the similarities,
          the dimensions of the logits matrix are multiplied by world-size along each axis (Trove models do not gather scores from other processes during eval).
        * In a distributed setup, the model vertically stacks all the gathered queries from all processes before calculating the scores. Then, Trainer
          also stacks the computed scores vertically once returned by the model. So, we need to make sure to account for these duplicate rows in a distributed environment.
          (we don't need to worry about this since Trove models do not do this if model is in eval mode (i.e., ``assert not model.training``)).

        one more thing to be aware of: on each process, Trainer takes the logits from the model output and labels from the input (i.e., without model touching them).
        And then processes them as needed (e.g., gather and concat from other processes, etc.).
        Just be aware that if you modify the labels in your model, that change is not reflected in the labels that this function receives.

        Args:
            eval_pred (EvalPrediction): evaluation results passed by transformers.Trainer
            compute_result (Optional[bool]): If true, return the average metrics over all queries seen so far instead of metrics for the current batch.
                And also, reset the state. I.e., clear the metrics gathered so far.

        Returns:
            The calculated IR metrics. A mapping from metric_name to metric_value
        """
        labels = eval_pred.label_ids
        logits = eval_pred.predictions

        # remove the duplicate rows in a distributed environment. See docstring for more info
        logits = logits[: len(labels)]

        if not isinstance(labels, torch.Tensor):
            labels = torch.tensor(labels)
        if not isinstance(logits, torch.Tensor):
            logits = torch.tensor(logits)

        # Extract the similarity docs between the query and its corresponding documents
        # if d_i_j is the j_th document for i_th query, the docs are ordered like:
        # [d_0_0, d_0_1, d_0_2, d_1_0, d_1_1, d_1_2, ...]
        # to select the subset of scores between queries and their corresponding documents, we create the following index
        # [[0, 1, 2],
        #  [3, 4, 5],
        #  ...
        #  ]
        # using these indices with torch.gather along dim 1 gives us the predicted sim scores

        num_queries, docs_per_query = labels.shape
        gather_idx = torch.arange(
            num_queries * docs_per_query, device=logits.device, dtype=torch.long
        )
        gather_idx = gather_idx.reshape(num_queries, docs_per_query)
        gather_idx = torch.remainder(gather_idx, logits.shape[1])
        logits = torch.gather(logits, dim=1, index=gather_idx)

        # <<<<<<<<<<<< alternative to the above operation that is slightly faster (very small difference) but more difficult to understand >>>>>>>
        # num_queries, docs_per_query = labels.shape
        # queries_per_batch = logits.shape[1] // docs_per_query
        # all_sub_logits = list()
        # for b_qidx in range(0, num_queries, queries_per_batch):
        #     sub_logits = logits[b_qidx: b_qidx + queries_per_batch]
        #     sub_logits = sub_logits.unfold(1, docs_per_query, docs_per_query).diagonal().T
        #     all_sub_logits.append(sub_logits)
        # logits = torch.vstack(all_sub_logits)
        # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        logits = logits.detach().cpu().tolist()
        labels = labels.detach().cpu().tolist()

        scores = defaultdict(dict)
        qrels = defaultdict(dict)

        for q_idx, (gt_labels, pred_scores) in enumerate(zip(labels, logits)):
            for doc_idx, (gt_label, pred_score) in enumerate(
                zip(gt_labels, pred_scores)
            ):
                scores[str(q_idx)][str(doc_idx)] = pred_score
                qrels[str(q_idx)][str(doc_idx)] = gt_label

        scores = dict(scores)
        qrels = dict(qrels)
        results = self.add_batch(scores=scores, qrels=qrels)

        if compute_result:
            # Average across all examples seen so far
            results = self.aggregate_results()
            self.reset_state()

        return results

    def check_training_arguments(self, args: TrainingArguments):
        """If used with ``transformers.Trainer``, check that we support the arguments that it
        uses."""
        if not args.batch_eval_metrics:
            raise NotImplementedError(
                "Calculating metrics for the aggregated prediction results is not supported."
            )



"""Helpers to deal with files and directories on disk.

Utilities in this file should be general and not limited to Trove. I.e., they should not contain
logic of tasks, etc. They are just convenience tools.
"""

import functools
import json
import os
import shutil
import stat
import weakref
from contextlib import contextmanager, nullcontext
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import uuid4

import huggingface_hub.utils as hub_utils
import xxhash
from datasets.utils._filelock import FileLock
from tqdm import tqdm
from transformers.utils import logging as hf_logging

from . import cache_manager, config
from .logging_utils import get_logger_with_config, rpath

logger, logging_conf = get_logger_with_config("trove")

hf_logger = hf_logging.get_logger(__name__)


class _TempPath:
    """A class representing a path that should be removed before exit.

    Adapted from `here <https://github.com/huggingface/datasets/blob/17f17b3fe7f276e1b019cca8aa651bf7c818a928/src/datasets/fingerprint.py#L45>`_ .

    You should **NOT** use instances of this class. You should use ``_TempPath.register("path/to/be/deleted")`` or
    ``_TempPath("path/to/be/deleted")`` and forget about it. We just need to create an object to be able
    to use `weakref` module to call cleanup functions.
    """

    # list of paths that are registered for removal
    _registered_paths: List = list()
    # keep a reference of instances of this class to be even more confident that they are
    # not garbage collected before reaching the end of the execution.
    _instances: List = list()

    @classmethod
    def register(cls, path: os.PathLike) -> None:
        """Register a path to be removed before exit if not already registered."""
        realp = realpath(path)
        if realp not in cls._registered_paths:
            cls._instances.append(cls(path))
            cls._registered_paths.append(realp)

    def __init__(self, path: os.PathLike):
        """Register a path to be removed before exit."""
        self.path = path
        # The finalizer is created within the object itself
        # and moreoever the callback function is a bound method of the object
        # itself. So, this object is never garbage collected and the cleanup is only called at
        # interpreter shutdown. This makes sure the directory is not removed while the code is still running.
        self._finalizer = weakref.finalize(self, self._cleanup)

    def _cleanup(self):
        """Remove the path if possible."""
        logger.debug(f"Remove path: {rpath(self.path)}")
        safe_remove_path(self.path)

    def cleanup(self):
        if self._finalizer.detach():
            self._cleanup()


def register_for_removal(path: os.PathLike):
    """Register path to be deleted before exit.

    .. warning::

        In distributed environments, call this function ONLY from one process and
        that should be the process that lasts the LONGEST. If the process that calls
        this function terminates before other processes, it will delete the registered path
        and if other running processes try to access it, they will either raise an exception or
        even worse, generate logically wrong results, which is hard to debug.

    Args:
        path: path to delete before the process exits.
    """
    _TempPath.register(path)


def write_qrels(
    qrels: Dict[str, Dict[str, Union[int, float]]],
    filepath: os.PathLike,
    format: str = "tsv",
) -> None:
    """Writes qrels to a file.

    Args:
        qrels: qrels in nested dict format. ``qrels[qid][docid]`` is the score (or annotation) between `qid` and `docid`.
        filepath: path to target file.
        format: Format of the target file. There are two options:
            * ``tsv`` is the standard qrel format in a tsv file with three columns ("query-id", "corpus-id", "score")
            * ``grouped`` is a json lines file where each record has three keys. ``qid`` is the ID of the current query.
              ``docid`` and ``score`` are two lists of the same size that hold the ID and score of the related documents for ``qid``
    """
    if format not in ["tsv", "grouped"]:
        msg = f"Possible formats for writing qrel records are 'tsv' and 'grouped'. Got: '{format}'"
        raise ValueError(msg)

    if format == "tsv":
        headers = ["query-id", "corpus-id", "score"]
        with CSVWriter(path=filepath, headers=headers, sep="\t") as writer:
            for qid, topdocs in qrels.items():
                for docid, score in topdocs.items():
                    writer.add_one((qid, docid, score))
    elif format == "grouped":
        with JSONLinesWriter(path=filepath) as writer:
            for qid, docdata in qrels.items():
                rec = {
                    "qid": qid,
                    "docid": list(docdata.keys()),
                    "score": list(docdata.values()),
                }
                writer.add_one(rec)


class CSVWriter:
    def __init__(
        self,
        path: os.PathLike,
        headers: Optional[List[str]] = None,
        sep: str = ",",
        chunk_size: Optional[int] = None,
    ) -> None:
        """Open a CSV file for writing.

        Example::

            with CSVWriter('/path/to/file.csv', headers=['qid', 'docid', 'score'], sep=',') as writer:
                writer.add([('a', 'b', 1), ('b', 'c', 2)])
                writer.add_one(('e', 'f', 3))

        Args:
            path: file to open.
            headers: a list of column names. If ``None``, file starts with the first row of data.
            sep: delimiter character
            chunk_size: flush the buffer every ``chunk_size`` records.
        """
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        self.fp = open(path, "w")
        self.sep = sep

        if headers is not None:
            self.fp.write(self.sep.join(headers) + "\n")

        if chunk_size is None:
            self.chunk_size = 1_000
        else:
            self.chunk_size = chunk_size
        self.line_buffer = list()

    def __enter__(self):
        """Act as a context manager."""
        return self

    def __exit__(self, *args, **kwargs):
        """Cleanups before exiting the context."""
        self.close()

    @staticmethod
    def write_to_file(
        records: Iterable[Any],
        path: os.PathLike,
        headers: Optional[List[str]] = None,
        sep: str = ",",
    ) -> None:
        """Write tuples to csv file.

        This methods opens the file, writes the records and closes the file.
        So, there is no incremental writing. It is mostly a convenience method for using
        ``CSVWriter`` when you have all the records (or a generator of records) ready before attempting to write them.

        Args:
            records: List/iterable of tuples to write to file.
            path: file to open.
            headers: a list of column names. If ``None``, file starts with the first row of data.
            sep: delimiter character
        """
        with CSVWriter(path=path, headers=headers, sep=sep) as writer:
            writer.add(records)

    def add(self, items: Iterable[Any]) -> None:
        """Write a list of tuples to file."""
        for item in items:
            self.line_buffer.append(self.sep.join([str(i) for i in item]))
            if len(self.line_buffer) >= self.chunk_size:
                self.flush()

    def add_one(self, item: Any) -> None:
        """Write one object to file."""
        self.add([item])

    def flush(self) -> None:
        """Flush the content of the buffer to file."""
        if len(self.line_buffer) != 0:
            self.fp.write("\n".join(self.line_buffer) + "\n")
            self.line_buffer = list()

    def close(self) -> None:
        """Flush the buffer and close the file."""
        self.flush()
        self.fp.close()


class JSONLinesWriter:
    def __init__(
        self, path: os.PathLike, chunk_size: Optional[int] = None, **kwargs
    ) -> None:
        """Open a json lines file for writing.

        Example::

            with JSONLinesWriter('/path/to/file.jsonl') as writer:
                writer.add([{'a': 1}, {'b': 2}])
                writer.add_one({'c': 3})

        Args:
            path: file to open.
            chunk_size: flush the buffer every ``chunk_size`` records.
            **kwargs: keyword arguments passed to ``json.dumps()``
        """
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        self.fp = open(path, "w")
        self.json_kw = kwargs
        if chunk_size is None:
            self.chunk_size = 1_000
        else:
            self.chunk_size = chunk_size
        self.line_buffer = list()

    def __enter__(self):
        """Act as a context manager."""
        return self

    def __exit__(self, *args, **kwargs):
        """Cleanups before exiting the context."""
        self.close()

    @staticmethod
    def write_to_file(records: Iterable[Any], path: os.PathLike, **kwargs) -> None:
        """Write records to json lines file.

        This methods opens the file, writes the records and closes the file.
        So, there is no incremental writing. It is mostly a convenience method for using
        ``JSONLinesWriter`` when you have all the records (or a generator of records) ready before attempting to write them.

        Args:
            records: List/iterable of objects to write to file.
            path: json lines file to write records to.
            **kwargs: keyword arguments passed to ``json.dumps()``
        """
        with JSONLinesWriter(path=path, **kwargs) as writer:
            writer.add(records)

    def add(self, items: Iterable[Any]) -> None:
        """Write a list of objects to file."""
        for item in items:
            self.line_buffer.append(json.dumps(item, **self.json_kw))
            if len(self.line_buffer) >= self.chunk_size:
                self.flush()

    def add_one(self, item: Any) -> None:
        """Write one object to file."""
        self.add([item])

    def flush(self) -> None:
        """Flush the content of the buffer to file."""
        if len(self.line_buffer) != 0:
            self.fp.write("\n".join(self.line_buffer) + "\n")
            self.line_buffer = list()

    def close(self) -> None:
        """Flush the buffer and close the file."""
        self.flush()
        self.fp.close()


def easyfilelock(path: os.PathLike, *args, **kwargs) -> FileLock:
    """Indirectly locks a file for writing.

    Instead of locking the main file, it creates a new lock file by
    adding ``.lock`` suffix to the main filepath and locks this new file instead.

    See why: https://py-filelock.readthedocs.io/en/latest/index.html

    Args:
        path (os.PathLike): filepath to lock for writing

    Returns:
        an instance of ``FileLock`` for writing to ``path``.
    """
    Path(path).parent.mkdir(exist_ok=True, parents=True)
    lock_file = Path(path).as_posix() + ".lock"
    return FileLock(lock_file, *args, **kwargs)


def realpath(path: os.PathLike) -> os.PathLike:
    """Return resolved absolute path.

    It is similar to linux ``realpath`` command. If ``path`` does exist, join realpath of its
    parent with its basename.
    """
    path = Path(path)
    if path.exists():
        return path.absolute().resolve().as_posix()
    else:
        basename = path.name
        dirname = path.parent.as_posix()
        return Path(realpath(dirname), basename).as_posix()


def _contains_symlink(path: os.PathLike) -> bool:
    """Check if a path is symlink or contains a symlink.

    Returns:
        True if the path itself is a symlink or it contains a symlink. False otherwise.
    """
    path = Path(path)
    if path.is_symlink():
        return True
    if path.is_dir():
        for subpath in path.iterdir():
            if _contains_symlink(subpath):
                return True
    # If we get to this point, it means the path is not a symlink and we didn't find a symlink
    # in any of its subdirectories. So return False
    return False


def safe_remove_path(path: os.PathLike, try_chmod: bool = False) -> None:
    """Try to remove a directory. Do not raise exceptions on failure.

    Adapted from `here <https://github.com/huggingface/huggingface_hub/blob/26c2d89f5e521b50916fbaf190e5a42705bcb775/src/huggingface_hub/utils/_fixes.py#L43>`_ .

    Args:
        path (os.PathLike): Path to delete.
        try_chmod (bool): When True, if failed to remove ``path``, try to change its permissions and try again.
    """
    _realpath = realpath(path)
    if not Path(_realpath).exists():
        # Nothing to remove
        return
    try:
        if Path(_realpath).is_dir():
            shutil.rmtree(_realpath)
        else:
            Path(_realpath).unlink()
    except:
        if try_chmod:
            # If failed, change permission and try again.
            try:
                os.chmod(_realpath, stat.S_IWRITE)
                if Path(_realpath).is_dir():
                    shutil.rmtree(_realpath)
                else:
                    Path(_realpath).unlink()
            except:
                pass


def call_cleanup(cleanup_fn: str):
    """A decorator to call a clean up function on class methods.

    It ensures to call ``cleanup_fn`` method of the object after the decorated
    function returns or raises an exception.

    Use like this::

        Class A:
            def mycleanup(self):
                ...

            @call_cleanup('mycleanup')
            def some_fn(self, ...):
                ...

    Args:
        cleanup_fn (str): The name of the class method to call.
    """

    def _clean_up_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            obj = args[0]
            try:
                return func(*args, **kwargs)
            finally:
                getattr(obj, cleanup_fn)()

        return wrapper

    return _clean_up_decorator


def hash_file_bytes_functional(
    path: os.PathLike, chunk_size: Optional[int] = None
) -> str:
    """Calculates a non-cryptographic hash of the file content.

    It repeats the calculation every time that it's called.
    Use :func:`hash_file_bytes` that relies on cached results to avoid duplicate computations.

    Args:
        path (os.PathLike): file to calculate the bash for
        chunk_size (Optional[int]): read file in chunks of ``chunk_size`` bytes. The default is 1MB.

    Returns:
        hash of the file content.
    """
    logger.debug("Hashing content of file.")
    if chunk_size is None:
        chunk_size = 2**20
    file_hash = xxhash.xxh64()
    num_bytes = 0
    with open(path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            num_bytes += len(chunk)
            file_hash.update(chunk)

    file_hash = file_hash.hexdigest()

    formatted_size = tqdm.format_sizeof(num_bytes, divisor=1024, suffix="B")
    logger.debug(f"Calculated hash for file with size: {formatted_size}")
    return file_hash


def hash_file_bytes(path: os.PathLike, chunk_size: Optional[int] = None) -> str:
    """Returns a non-cryptographic hash of the file content.

    It tries to read the hash from cache if it exists. If missing, calculate the hash and
    also write it to a cache file.

    Args:
        path (os.PathLike): file to calculate the hash for.
        chunk_size (Optional[int]): see :func:`hash_file_bytes_functional` for details.

    Returns:
        hash of the file content.
    """
    cache_file = cache_manager.get_cache_dir(
        input_data=path,
        artifact_type="final",
        artifact_content="content_hash",
        fingerprint_src="path",
    )
    cache_file = cache_file.joinpath("unique_hash_of_bytes.txt")
    with easyfilelock(cache_file.as_posix()):
        if not cache_file.exists():
            file_hash = hash_file_bytes_functional(path=path, chunk_size=chunk_size)
            with open(cache_file, "w") as f:
                f.write(file_hash)

    with open(cache_file, "r") as f:
        file_hash = f.read().strip()
    return file_hash


# from here:
# https://github.com/huggingface/huggingface_hub/blob/d30853b36029bd838500614cdae63b3bb55a8950/src/huggingface_hub/file_download.py#L1564
def _chmod_and_move(src: Path, dst: Path) -> None:
    """Set correct permission before moving a blob from tmp directory to cache dir.

    Do not take into account the ``umask`` from the process as there is no convenient way
    to get it that is thread-safe.

    See:
        - `About umask <https://docs.python.org/3/library/os.html#os.umask>`_
        - `Thread-safety <https://stackoverflow.com/a/70343066>`_
        - `About solution <https://github.com/huggingface/huggingface_hub/pull/1220#issuecomment-1326211591>`_
        - `Fix issue <https://github.com/huggingface/huggingface_hub/issues/1141>`_
        - `Fix issue <https://github.com/huggingface/huggingface_hub/issues/1215>`_
    """
    # Get umask by creating a temporary file in the cached repo folder.
    tmp_file = dst.parent.parent / f"tmp_{uuid4()}"
    try:
        tmp_file.touch()
        cache_dir_mode = Path(tmp_file).stat().st_mode
        os.chmod(str(src), stat.S_IMODE(cache_dir_mode))
    except OSError as e:
        hf_logger.warning(
            f"Could not set the permissions on the file '{src}'. "
            f"Error: {e}.\nContinuing without setting permissions."
        )
    finally:
        try:
            tmp_file.unlink()
        except OSError:
            # fails if `tmp_file.touch()` failed => do nothing
            # See https://github.com/huggingface/huggingface_hub/issues/2359
            pass

    shutil.move(str(src), str(dst), copy_function=_copy_no_matter_what)


# from here:
# https://github.com/huggingface/huggingface_hub/blob/d30853b36029bd838500614cdae63b3bb55a8950/src/huggingface_hub/file_download.py#L1599
def _copy_no_matter_what(src: str, dst: str) -> None:
    """Copy file from src to dst.

    If ``shutil.copy2`` fails, fallback to ``shutil.copyfile``.
    """
    try:
        # Copy file with metadata and permission
        # Can fail e.g. if dst is an S3 mount
        shutil.copy2(src, dst)
    except OSError:
        # Copy only file content
        shutil.copyfile(src, dst)


@contextmanager
def atomic_write(file: os.PathLike, root: str = "parent"):
    """Write to file only if the write operation is completed successfully.

    It returns a path to a temporary file that you should use instead of the original file path.
    If the write operation is completed successfully, it moves the temporary file to the original filepath.
    If write operation fails for any reason, it just removes the temporary file if it can (it does not raise errors if
    it fails to remove the temporary file).

    You should always work with the original filepath and only call this context manager
    at the very last moment before opening the file for writing. For example, if you need to check
    for patterns in the filepath or check the file extension, you should use the original filepath.

    Example::

        path = '/some/path/to/my/file.jsonl'
        with atomic_write(path) as new_path:
            with open(new_path, 'w') as f:
                f.write('some content')

    Keep in mind that the temporary filepath is totally random and changes everytime you call this function even with the same arguments.

    .. attention::

        If you are using ``trove.file_utils.easyfilelock()``, you must lock the original filepath and not the one returned by this function.

    Args:
        file: filepath to write the results to if the write operation is successful.
        root: where to create the temporary file. You can avoid copying the temporary file
            and just renaming it (which is much faster for large files) if you choose a
            directory that is on the same filesystem as the original filepath. Valid options are:
                - ``system``: use the system's temporary directory (i.e., wherever python's tempfile module creates its files).
                - ``cache``: use the central cache directory used in trove library
                - ``parent``: create a directory with specific name (name is the value
                    of ``trove.config.TROVE_INCOMPLETE_FILE_DIRNAME`` variable) in the parent directory of filepath.

    Yields:
        the temporary filepath that the user must use for write operations in this context manager.
    """
    valid_roots = ["system", "parent", "cache"]
    if root not in valid_roots:
        msg = f"Valid options for 'root' are '{valid_roots}'. Got: 'root': '{root}'"
        raise ValueError(msg)

    if root == "system":
        temp_dir_cm = hub_utils.SoftTemporaryDirectory()
    else:
        if root == "parent":
            # temp file will be in a subdir in file's parent
            _tr = Path(file).parent.joinpath(config.TROVE_INCOMPLETE_FILE_DIRNAME)
        elif root == "cache":
            # temp file will be in trove's centralized cache
            _tr = cache_manager.get_cache_pardir(
                artifact_content=config.TROVE_INCOMPLETE_FILE_DIRNAME,
                artifact_type="temp",
            )
        else:
            raise ValueError
        temp_dir_cm = nullcontext(Path(_tr).as_posix())

    with temp_dir_cm as temp_dir:
        # Make sure temp files are unique. Although not guaranteed in future versions, try to keep
        # the same base name as the original file so not to break libraries that rely on the extension of  file.
        temp_file = Path(temp_dir, uuid4().hex, Path(file).name).as_posix()
        # create the parent directory to avoid errors
        Path(temp_file).parent.mkdir(exist_ok=True, parents=True)
        try:
            logger.debug(
                f"Start write transaction. Original file: {rpath(file)} Temp file: {rpath(temp_file)}"
            )
            # make it a drop-in replacement
            if isinstance(file, Path):
                yield Path(temp_file)
            else:
                yield temp_file
            # If the caller has not created the temp_file, there is nothing to do.
            if Path(temp_file).exists():
                logger.debug(
                    f"Finish write transaction by moving temp file to actual destination of: {rpath(file)}"
                )
                # Suppress warning from Path.touch() by creating its parent directories
                Path(file).parent.mkdir(exist_ok=True, parents=True)
                # Move the temp file and then remove it
                _chmod_and_move(src=Path(temp_file), dst=Path(file))
        finally:
            logger.debug(
                f"After write transaction. Remove the temp file before exit: {rpath(temp_file)}"
            )
            # The temp file and auxiliary files should be removed whether we succeed or fail
            for _file in Path(temp_file).parent.iterdir():
                Path(_file).unlink(missing_ok=True)
            # Do not make a giant directory with a bunch of empty subdirectories
            # We can delete the dir safely since it is unique to this specific function call
            assert Path(temp_file).parent.exists() and not len(
                list(Path(temp_file).parent.iterdir())
            )
            Path(temp_file).parent.rmdir()


def create_file_name_to_id_mapping(
    files: Union[os.PathLike, List[os.PathLike]], id_length: str = "shortest"
) -> Dict[str, str]:
    """Create unique IDs for a list of files.

    The resulting file ID is a substring of the hexdigest of the hash of the binary content of the file.

    **NOTE: It will fail if a subset of files in ``files`` are copy of each other, i.e., have same hash.**

    Args:
        files (Union[os.PathLike, List[os.PathLike]]): a list of filepath to creates unique IDs for.
        id_length (str): The number of characters in the final ID. Acceptable values are

            * ``shortest``: Start from length=4 and increment the length by 1 until you get unique IDs for all files.
            * ``full``: use the full hash as ID.

    Returns:
        A mapping where keys are the elements of the ``files`` argument and the values are the corresponding unique file IDs.
    """
    if not isinstance(files, list):
        files = [files]

    # Ensure files are of 'str' dtype
    orig_files = [Path(p).as_posix() for p in files]
    realpath_files = [realpath(p) for p in orig_files]
    # the subset of unique files with their actual path (not symlinks, etc.)
    realpath_files = list(set(realpath_files))

    # Try to preserve the original IDs calculated by MaterializedQRel instances
    realpath_to_hash = dict()
    for _realpath in realpath_files:
        realpath_to_hash[_realpath] = hash_file_bytes(_realpath)

    if id_length == "shortest":
        _id_length = 4
    elif id_length == "full":
        _id_length = len(next(iter(realpath_to_hash.values())))
    else:
        msg = f"Possible values for 'id_length' are: 'shortest', 'full'. Got: '{id_length}'"
        raise ValueError(msg)

    while True:
        if _id_length > len(next(iter(realpath_to_hash.values()))):
            msg = "At least two of the files are identical copies of each other and it is not possible to make different IDs for them."
            raise RuntimeError(msg)

        realpath_to_id = {k: v[:_id_length] for k, v in realpath_to_hash.items()}
        if len(set(list(realpath_to_id.values()))) == len(realpath_files):
            break
        else:
            _id_length += 1

    # This block is never executed. It is remained here from previous versions.
    if len(set(list(realpath_to_id.values()))) != len(realpath_files):
        # If hashes are not unique, use sequential IDs
        realpath_to_id = {f: str(i) for i, f in enumerate(list(sorted(realpath_files)))}

    # Convert the mapping keys from resolved absolute path
    # to the exact path that is used by MaterializedQRel instances
    orig_name_to_id = {p: realpath_to_id[realpath(p)] for p in orig_files}
    return orig_name_to_id



import copy
import itertools
import json
import os
import pickle as pkl
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import datasets
import numpy as np
from datasets.fingerprint import Hasher
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from .. import cache_manager, containers, file_utils
from ..containers import MaterializedQRel, MaterializedQRelConfig
from ..file_utils import JSONLinesWriter
from ..logging_utils import get_logger_with_config, rpath
from .data_args import DataArguments
from .ir_encoding_dataset import EncodingDataset

logger, logging_conf = get_logger_with_config("trove")


EVAL_CACHE_FILE_NAMES = {
    "file_info_for_encoding": "file_info_for_encoding.json",
    "global_qrel": "qrels_with_global_id.pkl",
    "local_qrel": "qrels_with_local_id.pkl",
}


class MultiLevelDataset(Dataset):
    def __init__(
        self,
        data_args: DataArguments,
        format_query: Callable[[str, Optional[str]], str],
        format_passage: Callable[[str, str, Optional[str]], str],
        qrel_config: Optional[
            Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]
        ] = None,
        eval_cache_path: Optional[os.PathLike] = None,
        train_cache_path: Optional[os.PathLike] = None,
        data_args_overrides: Optional[Dict[str, Any]] = None,
        num_proc: Optional[int] = None,
    ) -> None:
        """IR training dataset with multiple levels of relevancy (supports more than two levels).

        * Collection of related documents for queries are created from one or more ``qrel_config``
        * If there are multiple collections and a passage shows up in multiple collections,
          the data from the last collection takes precedence
          (i.e., whatever record the object corresponding to
          ``qrel_config[-1]`` returns will be used for that specific passage)

        Args:
            data_args (DataArguments): general arguments for loading and processing the data
            format_query (Callable[[str, Optional[str]], str]): A callable that takes the query
                text and optionally the dataset name and returns the formatted query text for the modle.
            format_passage (Callable[[str, str, Optional[str]], str]): A callable that takes
                the passage text and title and dataset name and returns the formatted passage text for the model.
            data_args_overrides (Optional[Dict[str, Any]]): A mapping from a subset of ``DataArguments`` attribute names to
                their new values. These key values override the corresponding attributes in ``data_args`` argument.
                It is useful if you want to create multiple datasets from the same ``DataArguments`` instance
                but make small changes for each dataset without creating new ``DataArguments`` instances.
            qrel_config (Optional[Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]]):
                Config for one or more collections of queries, passages, and the relation between them.
                The combination of these collections will make up the content of this dataset.
            eval_cache_path (Optional[os.PathLike]): **DO NOT USE**. For internal operations only and not stable.
                If given, create a dataset only for evaluation from cache files in this directory.
                This is much more memory efficient compared to creating the dataset on-the-fly.
                You should use :meth:`export_and_load_eval_cache` method to take advantage of this.
            train_cache_path (Optional[os.PathLike]): **DO NOT USE**. For internal operations only and not stable.
                If given, create a dataset only for training from this cache file.
                This is much more memory efficient compared to creating the dataset on-the-fly.
                You should use :meth:`export_and_load_train_cache` method to take advantage of this.
            num_proc (Optional[int]): arg to to methods like ``datasets.Dataset.*``
        """
        super().__init__()
        # A unique fingerprint for the underlying data for this dataset.
        # Note that two datasets with the same fingerprint are only based on the same data
        # but can generate different samples. see docstring for self.fingerprint for details.
        # do not use this directly. Instead use 'self.fingerprint' attribute
        self._unique_fingerprint = None
        # some metadata about this dataset to be saved along with any cache files if necessary
        # Do not use this directly. Instead, use 'self.info' attribute
        self._info = None
        # If true, MultiLevelDataset[i] returns the training group (i.e., output of
        # self.create_group() which just returns the related docs) for each query rather than a training sample
        # this is useful for iterating over the underlying data for each dataset instance.
        self._iter_over_groups = False

        if (
            sum(i is not None for i in [qrel_config, eval_cache_path, train_cache_path])
            != 1
        ):
            msg = (
                "You should pass exactly on of the 'qrel_config', 'eval_cache_path', or 'train_cache_path' arguments."
                f" Got: 'qrel_config': '{qrel_config}', 'eval_cache_path': '{eval_cache_path}', 'train_cache_path': '{train_cache_path}'"
            )

            raise ValueError(msg)

        # make a deepcopy so we do not impact operations outside of this class.
        data_args = copy.deepcopy(data_args)
        if data_args_overrides is not None:
            for attr, val in data_args_overrides.items():
                if not hasattr(data_args, attr):
                    # We should not add new attributes
                    msg = f"'{attr}' is not a valid attribute name for 'DataArguments'"
                    raise ValueError(msg)
                setattr(data_args, attr, val)
        self.args = data_args

        if self.args.passage_selection_strategy not in [
            None,
            "random",
            "least_relevant",
            "most_relevant",
        ]:
            msg = (
                f"We only support [None, 'random', 'least_relevant', 'most_relevant'] for `passage_selection_strategy`."
                f" But, got: '{self.args.passage_selection_strategy}'"
            )
            raise ValueError(msg)

        self.format_query = format_query
        self.format_passage = format_passage

        # by default, the training groups should be created on-the-fly
        self.create_group = self.create_group_on_the_fly

        self.train_records = None
        self.qrel_with_global_ids_cache = None
        self.qrel_with_local_ids_cache = None
        self.file_info_for_encoding_cache = None
        if train_cache_path is not None:
            logger.info(
                f"Creating MultiLevelDataset from cached training groups in {rpath(train_cache_path)}."
            )
            if Path(train_cache_path).suffix != ".jsonl":
                msg = f"Only supported format for 'train_cache_path' is json lines. Got: {train_cache_path}"
                raise ValueError(msg)
            self.train_records = datasets.load_dataset(
                "json", data_files=Path(train_cache_path).as_posix(), split="train"
            )
            self.create_group = self.create_group_from_cache
            return
        if eval_cache_path is not None:
            logger.info(
                f"Creating MultiLevelDataset for evaluation from cached files in {rpath(eval_cache_path)}."
            )
            _cache_dir = Path(eval_cache_path)
            self.file_info_for_encoding_cache = (
                _cache_dir / EVAL_CACHE_FILE_NAMES["file_info_for_encoding"]
            )
            self.qrel_with_global_ids_cache = (
                _cache_dir / EVAL_CACHE_FILE_NAMES["global_qrel"]
            )
            self.qrel_with_local_ids_cache = (
                _cache_dir / EVAL_CACHE_FILE_NAMES["local_qrel"]
            )
            return

        if not isinstance(qrel_config, list):
            qrel_config = [qrel_config]

        # This is for logging purposes only. Do NOT use
        self._qrel_config = qrel_config

        _all_files = list()
        for conf in qrel_config:
            conf.ensure_list_of_correct_dtype()  # make sure everything is of type 'List'
            if conf.query_path is not None:
                if isinstance(conf.query_path, list):
                    _all_files.extend(conf.query_path)
                else:
                    _all_files.append(conf.query_path)

            if conf.corpus_path is not None:
                if isinstance(conf.corpus_path, list):
                    _all_files.extend(conf.corpus_path)
                else:
                    _all_files.append(conf.corpus_path)

        # Create a unique file id for each file path used in this dataset
        # This is used by MaterializedQRel instances to create globally unique query/passage IDs across files
        # which allows us to merge various files without concerns about duplicate IDs
        fname_to_id_map = file_utils.create_file_name_to_id_mapping(
            files=_all_files, id_length="shortest"
        )

        all_record_mappings = dict()
        for path in _all_files:
            _realpath = Path(path).resolve().absolute().resolve().as_posix()
            if _realpath not in all_record_mappings:
                all_record_mappings[_realpath] = containers.RowsByKeySingleSource(
                    filepath=path,
                    new_key_field="_id",
                    add_filepath_field=True,
                    num_proc=num_proc,
                )

        # Load the collections that make up this dataset.
        logger.info("Load all MaterializedQRel that make up this dataset.")
        self.qrel_collections = [
            MaterializedQRel(
                args=item,
                file_name_to_id=fname_to_id_map,
                record_mapping_collection=all_record_mappings,
                num_proc=num_proc,
            )
            for item in qrel_config
        ]

        all_qids = list(
            itertools.chain.from_iterable(
                [p.all_qids_global for p in self.qrel_collections]
            )
        )
        # ID of queries that are available in this dataset
        # Sort them for consistency across different runs
        self.all_qids = list(sorted(list(set(all_qids))))

    def update_metadata(self) -> None:
        """Updates the metadata for the dataset.

        It creates a new fingerprint and metadata dict for the dataset.
        """
        logger.debug("Creating dataset's fingerprint and info objects.")
        # The underlying data for each dataset is the set of MaterializedQRel collections
        # So its fingerprint is also the combined fingerprint of the MaterializedQRel collections that it contains.
        hasher = Hasher()
        for qrel_col in self.qrel_collections:
            hasher.update(qrel_col.fingerprint)

        # This is a workaround and not optimal. It should be changed in future.
        # There is some information in each MaterializedQRel that does not change
        # the state of the dataset and they are just there for ease of use and as
        # convenience tricks (e.g., MaterializedQRel.args.query_cache). Since changing them
        # does not change the state of the dataset, they ideally should not be included in dataset fingerprint.
        # However, if the user caches the dataset (with export_and_load_* methods) and then in a different run
        # they change these convenience fields and try to load the cache, the loaded dataset should contain the new values
        # for convenience fields.
        # In the final design, the convenience info should not be cached and instead their current value should be passed
        # to the cached datasets __init__ method when loading it in export_and_load_* methods.
        # Until then, we write and read them from cache and include them in the fingerprint so if they change,
        # the cache is invalidated and a new cache is created (which is not optimal.
        # All files in new cache will be the same except the convenience info files.)
        _encoding_path_info = self._get_encoding_path_info()
        for ftype_k in sorted(_encoding_path_info):
            for fpath_k in sorted(_encoding_path_info[ftype_k]):
                for field_k in sorted(_encoding_path_info[ftype_k][fpath_k]):
                    hasher.update(ftype_k)
                    hasher.update(fpath_k)
                    hasher.update(field_k)
                    hasher.update(_encoding_path_info[ftype_k][fpath_k][field_k])
        self._unique_fingerprint = hasher.hexdigest()
        self._info = {
            "fingerprint": self._unique_fingerprint,
            "data_args": self.args.to_dict(),
            "qrel_config": [qc.to_dict() for qc in self._qrel_config],
        }

    @property
    def fingerprint(self) -> str:
        """A unique fingerprint for the contents and output of this dataset.

        Datasets with the same fingerprint are backed by the same underlying data but do **NOT**
        necessarily generate the same samples. For example, using different sampling strategies
        or query and document formatting functions leads to different output from the same underlying
        data and thus the same fingerprint.

        This fingerprint is for internal operations only and you should not rely on it. And if you do,
        just use it to identify the underlying data (e.g., for caching and loading) and not
        the exact samples.

        datasets with the same fingerprint are backed by the same data sources.
        """
        if self._unique_fingerprint is None:
            self.update_metadata()
        return self._unique_fingerprint

    @property
    def info(self) -> Dict:
        if self._info is None or self._info["fingerprint"] != self.fingerprint:
            self.update_metadata()
        return self._info

    def set_index_lookup_storage_type(self, storage: str) -> None:
        """Select if the key to row index lookup table should be stored in memory or in memory-
        mapped lmdb dict."""
        for qc in self.qrel_collections:
            qc.set_index_lookup_storage_type(storage)

    def create_group_from_cache(self, index: int) -> Dict[str, Union[str, List[Dict]]]:
        """Loads one query and its related passages.

        Input and output is the same as :meth:`create_group_on_the_fly` but reads the data from
        cache file rather than creating it on-the-fly.
        """
        return self.train_records[index]

    def create_group_on_the_fly(self, index: int) -> Dict[str, Union[str, List[Dict]]]:
        """Loads one query and its related passages.

        The content of query/passages is also loaded (i.e., records are materialized.).

        If a passage exists in multiple collections, we use the data from the last collection
        in the self.qrel_collections list.

        Args:
            index (int): Index of the query to load

        Returns:

            A dict of the following format::

                {
                    'query_id': 'Unique ID of query across all files in this dataset'.
                    'query': 'query text',
                    'passages': [ # list of related passages for this query
                    {'_id': 'globally unique id of the passage', 'text': '...', 'title': '...'} # There could be additional fields in this dict, which should be ignored
                    ...,
                    {'_id': ....}
                    ]
                }
        """
        g_qid = self.all_qids[index]

        # Collect query/passage records in dicts from `_id` to `record` to avoid duplicate records
        query_rec = dict()
        doc_recs = dict()

        # Load related passages from all sources
        for qrel_col in self.qrel_collections:
            _q, _docs = qrel_col.get_related_recs(
                global_qid=g_qid, materialize=True, return_global_ids=True, strict=False
            )
            if _q is None or _docs is None:
                # this collection does not have entries for this specific query
                continue

            query_rec[_q["_id"]] = _q
            for doc in _docs:
                doc_recs[doc["_id"]] = doc

        # Make sure all collections return the same data for this query
        assert len(query_rec) == 1
        query_rec = list(query_rec.values())[0]
        doc_recs = list(doc_recs.values())

        group = {
            "query_id": query_rec["_id"],
            "query": query_rec["text"],
            "passages": doc_recs,
        }
        return group

    def __len__(self) -> int:
        if self.train_records is not None:
            return len(self.train_records)
        else:
            return len(self.all_qids)

    def __getitem__(
        self, index: int
    ) -> Dict[str, Union[str, List[str], List[Union[int, float]]]]:
        """Create a training example consisting of one query and a list of related passages.

        Returns:

            a dict with the following format::

                {
                    'query': formatted query of type `str`,
                    'passage': a list of formatted passages for the query.
                    'label': the relevancy level of each passage (i.e., score returned from qrel collections).
                }
        """
        group = self.create_group(index)
        if self._iter_over_groups:
            # we just need to return the training groups (and not training instances) without any processing
            return group
        query = group["query"]

        formatted_query = self.format_query(text=query, dataset=self.args.dataset_name)

        if self.args.group_size is None or self.args.group_size <= 0:
            # use all passages available
            passage_recs = group["passages"]
        else:
            # Sort the passage indices according to the given strategy
            # such that the index of the desired passages comes first
            # Also sort by '_id' field to make results consistent across different runs
            if self.args.passage_selection_strategy is None:
                curr_indices = list(range(len(group["passages"])))
            elif self.args.passage_selection_strategy == "random":
                _num_total = len(group["passages"])
                curr_indices = np.random.choice(
                    _num_total, [_num_total], replace=False
                ).tolist()
            elif self.args.passage_selection_strategy == "least_relevant":
                curr_indices = sorted(
                    range(len(group["passages"])),
                    key=lambda idx: (
                        group["passages"][idx]["score"],
                        group["passages"][idx]["_id"],
                    ),
                )
            elif self.args.passage_selection_strategy == "most_relevant":
                curr_indices = sorted(
                    range(len(group["passages"])),
                    key=lambda idx: (
                        group["passages"][idx]["score"],
                        group["passages"][idx]["_id"],
                    ),
                    reverse=True,
                )
            else:
                raise ValueError

            # choose the first self.args.group_size passages in the list
            # To also cover the case where slef.args.group_size > len(group['passages']),
            # we resample from the same passages again until reaching self.args.group_size
            chosen_idx = list(
                itertools.islice(itertools.cycle(curr_indices), self.args.group_size)
            )

            passage_recs = [group["passages"][i] for i in chosen_idx]

        formatted_passages = []
        passage_scores = list()
        for psg in passage_recs:
            formatted_passages.append(
                self.format_passage(
                    text=psg["text"],
                    title=psg["title"],
                    dataset=self.args.dataset_name,
                )
            )
            passage_scores.append(psg["score"])

        output = {
            "query": formatted_query,
            "passage": formatted_passages,
            "label": passage_scores,
        }
        return output

    def _get_encoding_path_info(self) -> Dict[str, Dict[str, Dict[str, str]]]:
        """Return the info about file paths needed to create encoding datasets.

        Returns:

            A dict with two keys: ``query_files`` and ``corpus_files``. Each value is also a dict
            from a file path to a mapping from field name to field values::

                {
                    'query_files':
                        {
                            '/path/to/file': {
                                'path': ...,
                                '_id': ...,
                                'orig_cache_path': ...,
                            }
                        }
                }
        """
        if self.file_info_for_encoding_cache is not None:
            # It is a cached eval dataset. So just return the cached values for the output of this method
            with open(self.file_info_for_encoding_cache, "r") as f:
                path_info_dict = json.load(f)
            return path_info_dict

        # Use mappings from absolute resolved path to file info
        # to collect a list of data files without duplicates.
        query_files = dict()
        corpus_files = dict()
        for qrel_col in self.qrel_collections:
            for path, _cache in zip(
                qrel_col.args.query_path, qrel_col.args.query_cache
            ):
                _path_info = {
                    "path": path,
                    "_id": qrel_col.file_name_to_id[path],
                    "orig_cache_path": _cache,
                }
                realpath = Path(path).absolute().resolve().as_posix()
                if realpath in query_files:
                    assert query_files[realpath]["_id"] == _path_info["_id"]
                query_files[realpath] = _path_info

            for path, _cache in zip(
                qrel_col.args.corpus_path, qrel_col.args.corpus_cache
            ):
                _path_info = {
                    "path": path,
                    "_id": qrel_col.file_name_to_id[path],
                    "orig_cache_path": _cache,
                }
                realpath = Path(path).absolute().resolve().as_posix()
                if realpath in corpus_files:
                    assert corpus_files[realpath]["_id"] == _path_info["_id"]
                corpus_files[realpath] = _path_info
        return {"corpus_files": corpus_files, "query_files": query_files}

    def get_encoding_datasets(
        self,
        get_cache_path: Optional[
            Callable[[os.PathLike, str, Optional[os.PathLike]], Optional[os.PathLike]]
        ] = None,
        encoding_cache_pardir: Optional[os.PathLike] = None,
        **kwargs,
    ) -> Tuple[List[EncodingDataset], List[EncodingDataset]]:
        """Generates encoding datasets for query and corpuses used in this dataset.

        This is most useful for inference and evaluation. You can use the datasets in the
        output of this method to calculate the query/corpus embeddings and then similarity scores between them.
        You can also use :meth:`get_qrel_nested_dict` to get the groundtruth qrels as a nested dict.
        With the groundtruth qrels and the calculated similarity scores, you can compute the IR evaluation metrics for the model.

        You can optionally assign new cache path to the encoding datasets.
        If you do so, the cache path in qrel collection arguments is ignored.
        You can use the arguments to this function to dynamically set the cache path for encoding datasets.

        Args:
            get_cache_path (Optional[Callable]): A callable that generates the cache path for each encoding dataset.
                The callable should take three keyword arguments:

                * filepath (`os.PathLike`):  path to the input data file
                * file_id (`str`): globally unique _id for this file (see code of ``__init__`` function for more info)
                * orig_cache_path (`Optional[PathLike]`): the corresponding cache path for this filepath saved in ``MaterializedQRel.args``

                And it should return ``None`` or the filepath to the cache for this dataset
            encoding_cache_pardir (Optional[os.PathLike]): If

                * ``get_cache_path`` is not provided (i.e., is ``None``)
                * and ``encoding_cache_pardir`` is not ``None``
                * and ``orig_cache_path`` is a relative filepath that does not exist on disk

                then, we assume ``orig_cache_path`` is a relative filepath and use
                ``Path(encoding_cache_pardir, orig_cache_path)`` as the cache path for the encoding dataset.
            kwargs: keyword arguments passed to `EncodingDataset.__init__`

        Returns:
            A tuple. The first item is a list of encoding datasets for queries used in this dataset
            The second item is a list of encoding datasets for corpuses used in this dataset.
        """
        logger.info(
            "Create EncodingDataset from query and corpus files used in this dataset."
        )

        if get_cache_path is None:

            def get_cache_path(
                filepath: os.PathLike,
                file_id: str,
                orig_cache_path: Optional[os.PathLike] = None,
            ) -> Optional[os.PathLike]:
                """Create the cache path for encoding datasets.

                See docstring for `get_encoding_datasets()` for details.
                """

                if encoding_cache_pardir is None or orig_cache_path is None:
                    return orig_cache_path
                if Path(orig_cache_path).root != "" or Path(orig_cache_path).exists():
                    return orig_cache_path
                return Path(encoding_cache_pardir, orig_cache_path).as_posix()

        _encoding_path_info = self._get_encoding_path_info()
        query_files = _encoding_path_info["query_files"]
        corpus_files = _encoding_path_info["corpus_files"]

        # instantiate the encoding datasets
        query_datasets = list()
        for path_info in query_files.values():
            query_datasets.append(
                EncodingDataset(
                    data_args=self.args,
                    query_path=path_info["path"],
                    format_query=self.format_query,
                    global_id_suffix=path_info["_id"],
                    cache_path=get_cache_path(
                        filepath=path_info["path"],
                        orig_cache_path=path_info["orig_cache_path"],
                        file_id=path_info["_id"],
                    ),
                    **kwargs,
                )
            )

        corpus_datasets = list()
        for path_info in corpus_files.values():
            corpus_datasets.append(
                EncodingDataset(
                    data_args=self.args,
                    corpus_path=path_info["path"],
                    format_passage=self.format_passage,
                    global_id_suffix=path_info["_id"],
                    cache_path=get_cache_path(
                        filepath=path_info["path"],
                        orig_cache_path=path_info["orig_cache_path"],
                        file_id=path_info["_id"],
                    ),
                    **kwargs,
                )
            )
        return query_datasets, corpus_datasets

    def get_qrel_nested_dict(
        self, return_global_ids: bool = True
    ) -> Dict[str, Dict[str, Union[int, float]]]:
        """Collect the qrel triplets from all qrel collections into the nested dict format used by
        ``pytrec_eval``.

        Args:
            return_global_ids (bool): if true, use global IDs for queries and documents.

        Returns:
            a nested dict where ``dict[qid][docid]`` is the score between query ``qid`` and document ``docid`` in qrel files.
        """
        logger.info("Create nested qrel dict for MultiLevelDataset instance.")

        if return_global_ids:
            cache_file = self.qrel_with_global_ids_cache
        else:
            cache_file = self.qrel_with_local_ids_cache

        if cache_file is not None:
            # It is a cached eval dataset. So, return the cached values for the output of this method.
            with open(cache_file, "rb") as f:
                qrel_dump = pkl.load(f)
            return qrel_dump

        qrel_mapping = defaultdict(dict)
        for qrel_col in self.qrel_collections:
            _qrel_dict = qrel_col.get_qrel_nested_dict(
                return_global_ids=return_global_ids
            )
            for qid, qdata in tqdm(
                _qrel_dict.items(),
                desc="Append Qrel Dict from one QRelCol",
                disable=not logging_conf.is_debug(),
            ):
                for docid, docscore in qdata.items():
                    qrel_mapping[qid][docid] = int(docscore)
        qrel_mapping = dict(qrel_mapping)
        return qrel_mapping

    def export_and_load_train_cache(
        self,
        cache_file: Optional[os.PathLike] = None,
        cache_pardir: Optional[os.PathLike] = None,
        num_proc: Optional[int] = None,
        batch_size: int = 16,
    ):
        """Export the training groups to a cache file and load them into a new instance of
        ``MultiLevelDataset``.

        To reduce memory consumption, it generates all the training groups, write them
        into a json lines file and returns a new ``MultiLevelDataset`` instance from those cached records.

        To benefit from the reduced memory consumption, make sure you do not keep any references to the old dataset instance,
        so it can be garbage collected by the interpreter. You can do something like::

            dataset = MultiLevelDataset(...)
            dataset = dataset.export_and_load_train_cache()
            gc.collect() # if you want to force the interpreter to release the memory right away

        Args:
            cache_file: a json lines files to save the cached training groups. If ``None``, a unique cache
                file is created based on the dataset fingerprint.
            cache_pardir: the directory to save the cache file to. If provided, we create a
                a subdir in dis directory based on the dataset fingerprint and save the dataset cache
                in that subdir.
            num_proc: number of workers to use to generate the training groups.
            batch_size: read the training groups in batches of this size

        Returns:
            A new instance of ``MultiLevelDataset`` for training that is backed by the cached training groups.
        """
        logger.info(
            "Export MultiLevelDataset training groups to cache file and load them in a new instance."
        )
        if num_proc is None:
            num_proc = 0

        if cache_file is not None and cache_pardir is not None:
            msg = (
                "You can pass at most one of the 'cache_file' or 'cache_pardir' arguments."
                f" Got: 'cache_pardir': '{cache_pardir}' and 'cache_file': '{cache_file}'"
            )
            raise ValueError(msg)

        if cache_file is None:
            if cache_pardir is not None:
                cache_file = cache_manager.get_cache_dir(
                    cache_pardir=cache_pardir,
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
            else:
                cache_file = cache_manager.get_cache_dir(
                    artifact_content="multilevel_dataset_train_cache",
                    artifact_type="final",
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
            cache_file = Path(cache_file, "train_records.jsonl")
        else:
            cache_file = Path(cache_file)
        cache_file.parent.mkdir(exist_ok=True, parents=True)

        logger.debug("Waiting to acquire the cache file lock.")
        with file_utils.easyfilelock(cache_file.as_posix()):
            if not cache_file.exists():
                logger.debug("Dataset cache file does not exist. Start creating it.")
                # We don't want to save the exact output of the dataset since that depends on sampling strategy, etc.
                # Those operations must be done during training and do not have a noticeable impact memory consumption
                orig_iter_groups = self._iter_over_groups
                self._iter_over_groups = True

                # use torch data loaders parallelize and speedup the read operations
                data_loader = DataLoader(
                    self,
                    batch_size=batch_size,
                    collate_fn=lambda x: x,
                    num_workers=num_proc,
                    drop_last=False,
                    shuffle=False,
                )
                with file_utils.atomic_write(file=cache_file, root="parent") as tfile:
                    with JSONLinesWriter(tfile) as rec_writer:
                        for batch in tqdm(
                            data_loader, desc="Export train groups to jsonl file"
                        ):
                            rec_writer.add(batch)

                # revert the dataset to its original format/functionality
                self._iter_over_groups = orig_iter_groups
        logger.debug("Instantiate a new instance of the dataset from the cache files.")
        # do not forget to deepcopy attributes that are part of this instance (and not references to other objects)
        # If you do not do so, this instance will not get garbage collected, which makes the entire caching pointless
        return MultiLevelDataset(
            data_args=copy.deepcopy(self.args),
            format_query=self.format_query,
            format_passage=self.format_passage,
            train_cache_path=cache_file,
            num_proc=num_proc,
        )

    def export_and_load_eval_cache(
        self,
        cache_dir: Optional[os.PathLike] = None,
        cache_pardir: Optional[os.PathLike] = None,
    ):
        """Export the data required for evaluation to cache files and load them into a new instance
        of ``MultiLevelDataset``.

        To reduce memory consumption, it creates all the data that is required for evaluation
        and writes them into cache files and returns a new ``MultiLevelDataset`` instance from those cache files.

        To benefit from the reduced memory consumption, make sure you do not keep any references to the old dataset instance,
        so it can be garbage collected by the interpreter. You can do something like::

            dataset = MultiLevelDataset(...)
            dataset = dataset.export_and_load_eval_cache()
            gc.collect() # if you want to force the interpreter to release the memory right away

        Args:
            cache_dir: a directory where cache files should be saved. If ``None``, create a unique
                cache directory based on the dataset fingerprint.
            cache_pardir: the parent directory to save the cache file to. If provided, we create a
                a subdir in this directory based on the dataset fingerprint and save the dataset cache
                in that subdir.

        Returns:
            A new instance of ``MultiLevelDataset`` for evaluation that is backed by the cached files.
        """
        logger.info("Write MultiLevelDataset evaluation data to cache files.")

        if cache_dir is not None and cache_pardir is not None:
            msg = (
                "You can pass at most one of the 'cache_dir' or 'cache_pardir' arguments."
                f" Got: 'cache_pardir': '{cache_pardir}' and 'cache_dir': '{cache_dir}'"
            )
            raise ValueError(msg)

        if cache_dir is None:
            if cache_pardir is not None:
                cache_dir = cache_manager.get_cache_dir(
                    cache_pardir=cache_pardir,
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
            else:
                cache_dir = cache_manager.get_cache_dir(
                    artifact_content="multilevel_dataset_eval_cache",
                    artifact_type="final",
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
        else:
            cache_dir = Path(cache_dir)
        cache_dir.mkdir(exist_ok=True, parents=True)

        # we only need these three files for evaluation
        path_info_file = cache_dir / EVAL_CACHE_FILE_NAMES["file_info_for_encoding"]
        global_qrel_file = cache_dir / EVAL_CACHE_FILE_NAMES["global_qrel"]
        # local_qrel_file = cache_dir / EVAL_CACHE_FILE_NAMES["local_qrel"] # ignore this for the pre-release version

        # a successful caching requires all these files to be written without any corruption
        # so lock them all (i.e., count them as one giant super file)
        with (
            file_utils.easyfilelock(path_info_file),
            file_utils.easyfilelock(global_qrel_file),
            # file_utils.easyfilelock(local_qrel_file), # ignore this for the pre-release version
        ):
            # if any of the files does not exist, then the cache is not valid and should be overwritten
            if not (
                path_info_file.exists()
                and global_qrel_file.exists()
                # and local_qrel_file.exists() # ignore this for the pre-release version
            ):
                _encoding_path_info = self._get_encoding_path_info()
                _global_qrel = self.get_qrel_nested_dict(return_global_ids=True)
                # _local_qrel = self.get_qrel_nested_dict(return_global_ids=False)

                # do an atomic_write operation for all files at the same time so if
                # only one of them fails to write, all the rest will also fail
                with (
                    file_utils.atomic_write(
                        file=path_info_file, root="parent"
                    ) as path_f,
                    file_utils.atomic_write(
                        file=global_qrel_file, root="parent"
                    ) as gbl_f,
                    # file_utils.atomic_write(file=local_qrel_file, root="parent") as lcl_f, # ignore this for the pre-release version
                ):
                    with open(path_f, "w") as f:
                        json.dump(_encoding_path_info, f, indent=2)
                    with open(gbl_f, "wb") as f:
                        pkl.dump(_global_qrel, f)
                    # ignore this for the pre-release version
                    # with open(lcl_f, "wb") as f:
                    #     pkl.dump(_local_qrel, f)
        # do not forget to deepcopy attributes that are part of this instance (and not references to other objects)
        # If you do not do so, this instance will not get garbage collected, which makes the entire caching pointless
        return MultiLevelDataset(
            data_args=copy.deepcopy(self.args),
            format_query=self.format_query,
            format_passage=self.format_passage,
            eval_cache_path=cache_dir,
        )



"""Utilities for transforming data structure and format with support for caching."""

import os
from pathlib import Path
from typing import Callable, Optional

import datasets

from .. import cache_manager, file_utils
from ..logging_utils import get_logger_with_config
from . import polars_transforms

logger, logging_conf = get_logger_with_config("trove")


def _group_qrel_triplets(
    src_file: os.PathLike,
    dst_file: os.PathLike,
    groupby_op: Callable[[os.PathLike, os.PathLike], None],
    num_proc: Optional[int] = None,
) -> None:
    """Read qrel triplets from file and Group them by qid field.

    At the moment, we only support reading qrel triplets from CSV files or infer them from sydir docids.

    We use ``groupby_op`` to read qrels and group them.
    Additionally, we take the output of these functions and cast qids and docids to ``str``.

    Args:
        src_file (os.PathLike): File to read or infer qrels triplets from.
        dst_file (os.PathLike): file to write the resulting grouped qrels to.
        groupby_op (Callable[[os.PathLike, os.PathLike], None]): a callable that does
            the actual groupby operation. It should take two keyword arguments, ``src_file``
            and ``dst_file``. It should read the qrels from ``src_file`` and write the grouped
            qrels to ``dst_file``.
        num_proc (Optional[int]): arg to ``datasets.Dataset.*`` methods
    """
    dst_file = Path(dst_file)
    if dst_file.suffix != ".jsonl":
        msg = f"We only support writing to jsonl file. But, got: {dst_file.as_posix()}"
        raise ValueError(msg)

    with file_utils.easyfilelock(dst_file):
        if not Path(dst_file).exists():
            # if dst_file does not exist, we should first use groupby_op to create a jsonl
            # file containing the output and then cast data types of that output if necessary
            dst_cache = cache_manager.get_cache_dir(
                input_data=src_file,
                artifact_content="grouped_qrel_triplets_polar",
                artifact_type="intermediate",
                fingerprint_src="path",
            )
            # output from groupby_op which might contain wrong data types
            dst_cache = Path(dst_cache, "grouped_qrels_pl.jsonl").as_posix()
            # we do not need to lock this file for writing because it is only written to
            # in this function and while the lock for dst_cache is acquired.
            # We do NOT need an atomic_write context for groupby_op it is always overwritten if dst_file
            # does not exists. And if dst_file exists, we successfully created the final file
            # and do not care if ouptut of groupby_op is corrupted later or not
            groupby_op(src_file=src_file, dst_file=dst_cache)

            logger.debug(
                "Load grouped qrels by HF dataset to potentially cast dtypes and save output as json lines."
            )
            # Read polars results with hf datasets to cast qid and docid if necessary and write the final results
            ds = datasets.load_dataset("json", data_files=dst_cache, split="train")

            # Make sure the IDs are of type 'str'
            new_features = ds.features.copy()
            if new_features["qid"] != datasets.Value("string") or new_features[
                "docid"
            ].feature != datasets.Value("string"):
                logger.debug("Cast 'qid' and 'docids' to 'str'")

                new_features["qid"] = datasets.Value("string")
                new_features["docid"].feature = datasets.Value("string")

                ds = ds.cast(new_features, num_proc=num_proc)

            with file_utils.atomic_write(file=dst_file, root="parent") as tfile:
                # Write results as json lines file
                ds.to_json(tfile)


def group_qrel_triplets_from_csv(
    src_file: os.PathLike,
    dst_file: os.PathLike,
    num_proc: Optional[int] = None,
) -> None:
    """Read qrel triplets from csv file and write them in grouped qrel format.

    see :func:`~trove.data.polars_transforms.group_qrel_triplets_from_csv_op` for the expected format
    of ``src_file`` and the format of the output ``dst_file``.

    Args:
        src_file (os.PathLike): csv/tsv file to read qrels from.
        dst_file (os.PathLike): output file to write the resulting grouped qrels
        num_proc (Optional[int]): max number of parallel process to use for data processing
    """
    _group_qrel_triplets(
        src_file=src_file,
        dst_file=dst_file,
        groupby_op=polars_transforms.group_qrel_triplets_from_csv_op,
        num_proc=num_proc,
    )


def group_qrel_triplets_from_sydir_corpus(
    src_file: os.PathLike,
    dst_file: os.PathLike,
    num_proc: Optional[int] = None,
) -> None:
    """Create qrel triplets from sydir document IDs and write them in grouped qrel format.

    see :func:`~trove.data.polars_transforms.group_qrel_triplets_from_sydir_corpus_op` for the expected format
    of ``src_file`` and the format of the output ``dst_file``.

    Args:
        src_file (os.PathLike): sydir corpus file with document IDs to use to create qrel triplets
        dst_file (os.PathLike): output file to write the resulting grouped qrels
        num_proc (Optional[int]): max number of parallel process to use for data processing
    """
    _group_qrel_triplets(
        src_file=src_file,
        dst_file=dst_file,
        groupby_op=polars_transforms.group_qrel_triplets_from_sydir_corpus_op,
        num_proc=num_proc,
    )



import os
import pickle as pkl
from pathlib import Path
from typing import Any, Dict, List, Optional

import datasets
from tqdm import tqdm

from .. import cache_manager, file_utils
from ..logging_utils import get_logger_with_config
from . import container_utils
from .lmdb_dict import LMDBDict

logger, logging_conf = get_logger_with_config("trove")


def encode_key(key: str) -> bytes:
    """Encode str keys as bytes for saving to LMDBDict."""
    return key.encode("utf-8")


def decode_key(key: bytes) -> str:
    """Decode keys read from LMDBDict from bytes to str."""
    return key.decode("utf-8")


def encode_value(value: List[int]) -> bytes:
    """Encode values to bytes using python pickle."""
    return pkl.dumps(value)


def decode_value(value: bytes) -> List[int]:
    """Decode values from bytes to original objects using python pickle."""
    return pkl.loads(value)


def lmdb_key_to_row_indices_from_dict(
    db_file: os.PathLike,
    db_hash_file: Optional[os.PathLike] = None,
    k2r_idx: Optional[Dict[str, List[int]]] = None,
    k2r_idx_file: Optional[os.PathLike] = None,
) -> LMDBDict:
    if (k2r_idx is None) == (k2r_idx_file is None):
        msg = (
            "You should pass exactly one of the 'k2r_idx' or 'k2r_idx_file' arguments."
            f" Got: type(k2r_idx): '{type(k2r_idx)}' and type(k2r_idx_file): '{type(k2r_idx_file)}'"
        )
        raise ValueError(msg)

    if db_hash_file is not None:
        db_hash_file = Path(db_hash_file)

    with file_utils.easyfilelock(db_file):
        # only process should write to db file. See LMDBDict docstrings for details.
        if not Path(db_file).exists():
            if k2r_idx is None:
                with open(k2r_idx_file, "rb") as f:
                    k2r_idx = pkl.load(f)
            hash_content = LMDBDict.create_db(
                data=k2r_idx.items(),
                db_file=db_file,
                encode_key=encode_key,
                encode_value=encode_value,
            )
            db_hash_file.parent.mkdir(exist_ok=True, parents=True)
            db_hash_file.write_text(hash_content)

    _hash = None
    if db_hash_file is not None and db_hash_file.exists():
        _hash = db_hash_file.read_text().strip()
    memmap_k2r_idx = LMDBDict.load_db(
        db_file=db_file,
        db_hash=_hash,
        encode_key=encode_key,
        decode_key=decode_key,
        encode_value=encode_value,
        decode_value=decode_value,
    )
    return memmap_k2r_idx


class KeyToRowIndices:
    def __init__(
        self,
        dataset: datasets.Dataset,
        id_field: str,
        storage: str = "in_memory",
        num_proc: Optional[int] = None,
    ) -> None:
        """Dict from key values to row indices for HF datasets.

        This class allows you to access HF dataset rows with the value of a specific column.

        It can use a memory-mapped dictionary implemented using ``lmdb`` to reduce memory usage.
        You do not need to use memory mapping. You should be fine with regular dictionary even for tens of millions of records.

        Example:

        .. code-block:: python

            ds = datasets.Dataset.from_dict({'_id': ['a', 'b'], 'val': ['foo', 'bar']})
            k2ridx = KeyToRowIndices(ds, id_field='_id')
            assert k2ridx['a'] == [0] # list of index rows that their '_id' field is equal to 'a'
            assert [ds[i] for i in k2ridx['a']] == [{'_id': 'a', 'val': 'foo'}]

        This class provides a subset of python dictionary API.

        Args:
            dataset (datasets.Dataset): dataset to create index mapping for.
            id_field (str): name of the field to use as key.
            storage (str): whether load the key to row indices mapping in memory or use
                a memory mapped dict (using ``lmdb``).
            num_proc (Optional[int]): arg to `datasets.Dataset.*` methods.
        """
        logger.debug(
            f"Create memory mapped key to row indices mapping. num records: {tqdm.format_sizeof(len(dataset))}"
        )

        if storage not in ["in_memory", "memory_mapped"]:
            msg = f"valid values for 'storage' argument are 'in_memory' and 'memory_mapped'. Got: '{storage}'"
            raise ValueError(msg)

        self.k2r_idx_pkl_file = container_utils.dataset_column_to_row_indices_mapping(
            dataset=dataset, id_field=id_field, num_proc=num_proc, load=False
        )
        db_pardir = cache_manager.get_cache_dir(
            input_data=dataset,
            artifact_type="final",
            artifact_content="lmdb_dict_key_to_row_indices_mapping",
        )
        self.k2r_idx_db_file = Path(db_pardir, "key_to_row_indices.db").as_posix()
        self.k2r_idx_db_hash_file = Path(db_pardir, "key_to_row_indices.hash")
        self.memmapped_key_to_row_idx = None

        if storage == "in_memory":
            with open(self.k2r_idx_pkl_file, "rb") as f:
                self.key_to_row_idx = pkl.load(f)
        elif storage == "memory_mapped":
            self.memmapped_key_to_row_idx = lmdb_key_to_row_indices_from_dict(
                db_file=self.k2r_idx_db_file,
                db_hash_file=self.k2r_idx_db_hash_file,
                k2r_idx_file=self.k2r_idx_pkl_file,
            )
            self.key_to_row_idx = self.memmapped_key_to_row_idx
        else:
            raise ValueError

    def close(self):
        """Close LMDB environment."""
        if self.memmapped_key_to_row_idx is not None:
            self.memmapped_key_to_row_idx.close()

    def __enter__(self):
        """Act as a context manager."""
        return self

    def __exit__(self, *args, **kwargs):
        """Cleanups before exiting the context."""
        self.close()

    def set_storage_type(self, storage: str) -> None:
        """Select if the mapping should be kept in memory or saved on disk and memory mapped using
        lmdb.

        Args:
            storage: where to save the data. Acceptable options are ``memory_mapped`` and ``in_memory``.
        """
        if storage == "memory_mapped":
            if self.memmapped_key_to_row_idx is None:
                self.memmapped_key_to_row_idx = lmdb_key_to_row_indices_from_dict(
                    db_file=self.k2r_idx_db_file,
                    db_hash_file=self.k2r_idx_db_hash_file,
                    k2r_idx=self.key_to_row_idx,
                )
            self.key_to_row_idx = self.memmapped_key_to_row_idx
        elif storage == "in_memory":
            if not isinstance(self.key_to_row_idx, dict):
                with open(self.k2r_idx_pkl_file, "rb") as f:
                    self.key_to_row_idx = pkl.load(f)
        else:
            msg = f"Only supported storage formats are 'memory_mapped' and 'in_memory'. Got: '{storage}'"
            raise ValueError(msg)

    def get(self, key: str, default: Optional[Any]) -> Any:
        """Similar to python's ``dict.get`` method."""
        val = self.key_to_row_idx.get(key, None)
        if val is None:
            return default
        else:
            return val

    def keys(self):
        """Similar to python's ``dict.keys``."""
        for key in self.key_to_row_idx.keys():
            yield key

    def values(self):
        """Similar to python's ``dict.values``."""
        for val in self.key_to_row_idx.values():
            yield val

    def items(self):
        """Similar to python's ``dict.items``."""
        for key, val in self.key_to_row_idx.items():
            yield (key, val)

    def __len__(self) -> int:
        """Similar to python's ``len(dict)``."""
        return len(self.key_to_row_idx)

    def __contains__(self, key: str) -> bool:
        """Support python's ``in`` operator like ``key in dict``."""
        return key in self.key_to_row_idx

    def __iter__(self):
        """Support iterating directly over ``KeyToRowIndices`` instances similar to python
        dictionaries."""
        return self.keys()

    def __getitem__(self, key: str) -> List[int]:
        """support indexing operator."""
        return self.key_to_row_idx[key]



import itertools
import math
import os
import pickle as pkl
from collections import defaultdict
from contextlib import ExitStack
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import datasets
import numpy as np
from datasets.fingerprint import Hasher
from tqdm import tqdm

from .. import cache_manager, file_utils
from ..data import file_reader
from ..logging_utils import get_logger_with_config
from . import container_utils, rows_by_key
from .key_to_row_indices import KeyToRowIndices
from .materialized_qrel_config import MaterializedQRelConfig
from .rows_by_key import RowsByKeySingleSource

logger, logging_conf = get_logger_with_config("trove")


class MaterializedQRel:
    def __init__(
        self,
        args: MaterializedQRelConfig,
        file_name_to_id: Optional[Dict[os.PathLike, str]] = None,
        record_mapping_collection: Optional[Dict[str, RowsByKeySingleSource]] = None,
        num_proc: Optional[int] = None,
    ) -> None:
        """Represents a collection of query and documents and the relation between them.

        You can use this class to combine multiple query files with multiple corpus files and represent the relation between them with a combination of multiple qrel files.

        For the collection of queries:

            * Each query should have a unique ID
            * Each query should appear only once in each file
            * Each query should only appear in one file
            * In sum, each query ID should only appear once across all your files.

        At the moment, we do not impose these restrictions strictly as it is computationally expensive for large collections.
        You will get logically wrong results (or in consistent, at best) if your files do not follow these restrictions.

        The same is true for the collection of corpus files and the collection of qrel files.
        For qrel files, each record should be unique with respect to the combination of ``qid`` and ``docid``.
        I.e, each (``qid``, ``docid``) combination should only appear once across all your qrel files.


        If you do not specify a qrel file (i.e., ``args.qrel_path`` is empty) You can
        still use this class for further organization as a namespace container to
        hold a query and a corpus collection (or just one of them) without any
        known relation between them. E.g. hold a query and corpus collection for hard negative mining
        without knowing the relevance levels between the two. Another example is a ``MaterializedQRel`` instance
        that only has a collection of documents (without qrels or even queries). You can mix such an instance with other MaterializedQRel
        instances in a :py:class:`trove.data.ir_dataset_multilevel.MultiLevelDataset` instance to expand the document pool during nearest neighbor search
        without impacting the qrel triplets (without adding new qrel triplets).

        Args:
            args (MaterializedQRelConfig): Information about files that contain the raw data and how to process them.
            file_name_to_id (Optional[Dict[os.PathLike, str]]): A mapping from file name to a unique ID (of type ``str``) for that file.
                Although each ``qid`` and ``docid`` is unique in each instance of ``MaterializedQRel``,
                you might want to combine multiple instances of ``MaterializedQRel`` that could potentially assign the same ``qid`` or ``docid`` to different examples.
                To make that possible and uniquely identify each record across files, we can create a new ID by combining the original ID (which is unique in each file)
                with a suffix that uniquely identifies the file that contains the record with that ID.
                ``file_name_to_id`` is the mapping from file names to fild IDs that are used for this purposes.
                if not provided, we use create this mapping based on the hash of the file bytes.
                You can update the file IDs later, but that leads to repeating a lot of the computations.
            record_mapping_collection (Optional[Dict[str, RowsByKeySingleSource]]): instance of :py:class:`trove.data.ir_dataset_multilevel.RowsByKeySingleSource` for query and corpus files.
                If provided, use these instances instead of loading the data into new instances.
            num_proc (Optional[int]): passed to `datasets.Dataset.*` methods.
        """
        # A unique fingerprint for the contents and output of this collection
        # do not use this directly. Instead use 'self.fingerprint' attribute
        self._unique_fingerprint = None
        # some metadata about this container to be saved along with any cache files if necessary
        # Do not use this directly. Instead, use 'self.info' attribute
        self._info = None
        # parent directory where cache files should be saved. Do not use directly.
        # Instead, use self.cache_dir
        self._cache_dir = None

        args.ensure_list_of_correct_dtype()
        self.args = args

        if file_name_to_id is not None:
            # Since different paths might refer to the same file (e.g., symlinks),
            # use the resolved absolute path (similar to linux realpath command) to refer to files
            _fn2id_rp = {file_utils.realpath(k): v for k, v in file_name_to_id.items()}
            # Find the new file IDs for files used in this class.
            self.file_name_to_id = dict()
            for key in list(set(self.args.query_path + self.args.corpus_path)):
                self.file_name_to_id[key] = _fn2id_rp[file_utils.realpath(key)]
        else:
            # Since we are sure there are no duplicate files, we can
            # use the returned name_to_id mapping as self.file_name_to_id directly
            self.file_name_to_id = file_utils.create_file_name_to_id_mapping(
                files=list(set(self.args.query_path + self.args.corpus_path)),
                id_length="shortest",
            )
        # Make sure the file IDs are unique
        assert len(self.file_name_to_id) == len(
            set(list(self.file_name_to_id.values()))
        )

        # If provided, qrel triplets are limited to this subset of queries
        qid_subset = set()
        if self.args.query_subset_path is not None:
            if not isinstance(self.args.query_subset_path, list):
                self.args.query_subset_path = [self.args.query_subset_path]
            for path in self.args.query_subset_path:
                qid_subset.update(
                    set(file_reader.load_qids(filepath=path, num_proc=num_proc))
                )
        qid_subset = sorted(list(qid_subset))

        logger.debug("Read qrel files to create MaterializedQRel's qrel_ds.")
        # List of datasets.Dataset, each with (qid, docid, score) columns.
        qrel_ds = list()
        for path in self.args.qrel_path:
            _subqrel = file_reader.load_qrel(filepath=path, num_proc=num_proc)
            if len(qid_subset):
                _subidx_mapping = KeyToRowIndices(
                    dataset=_subqrel, id_field="qid", num_proc=num_proc
                )
                _subidx = list()
                for _sid in qid_subset:
                    _subidx.append(_subidx_mapping.get(_sid, []))
                _subidx = itertools.chain.from_iterable(_subidx)
                _subqrel = _subqrel.select(_subidx)
            qrel_ds.append(_subqrel)

        if len(qrel_ds) == 0:
            qrel_ds = None
        elif len(qrel_ds) == 1:
            qrel_ds = qrel_ds[0]
        else:
            qrel_ds = datasets.concatenate_datasets(qrel_ds)

        user_filter_fn = self.args.filter_fn
        if user_filter_fn is None and (
            self.args.min_score is not None or self.args.max_score is not None
        ):
            user_filter_fn = lambda ex: (
                (self.args.min_score is None) or (self.args.min_score <= ex["score"])
            ) and ((self.args.max_score is None) or (ex["score"] < self.args.max_score))

        if user_filter_fn is not None and qrel_ds is not None:
            logger.debug(f"Filtering MaterializedQRel's qrel_ds")
            # In batched .map, if a group ends up being empty, it will be removed, which is equivalent to filtering
            # Why not use .filter directly? with that we can either keep/filter a group
            # and not able to select a subset of a group
            qrel_ds = qrel_ds.map(
                container_utils.filter_fn_wrapper(user_filter_fn),
                remove_columns=qrel_ds.column_names,
                batched=True,
            )

        # Collection of all qrel triplets that could be easily filtered by 'qid'
        self.qrel_mapping = rows_by_key.RowsByKey(
            dataset=qrel_ds, key_field="qid", num_proc=num_proc
        )

        # Score transformation is applied to query-passage scores before returning them
        # If you define score transform function directly in the body of the __init__ method
        # then whenever you want to calculate the hash of the class (and thus hash of self.score_transform),
        # you pickle the entire closure of self.score_transform (which is __init__ local scope)
        # That is huge, takes a lot of resources and time. So define the function in a separate
        # method to avoid that issue
        self.score_transform = MaterializedQRel.create_score_transform(
            args_score_transform=self.args.score_transform
        )

        # see above comments for score transform for why we create the group_filter_fn
        # in a separate function
        # see docs for MaterializedQRelConfig for signature of group_filter_fn
        self.group_filter_fn = MaterializedQRel.create_group_filter_fn(
            args_group_filter_fn=self.args.group_filter_fn,
            group_top_k=self.args.group_top_k,
            group_bottom_k=self.args.group_bottom_k,
            group_first_k=self.args.group_first_k,
            group_random_k=self.args.group_random_k,
        )

        if record_mapping_collection is not None:
            # Since different paths might refer to the same file (e.g., symlinks),
            # use the resolved absolute path (similar to linux realpath command) to refer to files
            rec_map_col_realpath = {
                Path(k).resolve().absolute().resolve().as_posix(): v
                for k, v in record_mapping_collection.items()
            }
            corpus_rec_maps = list()
            for path in self.args.corpus_path:
                _realpath = Path(path).resolve().absolute().resolve().as_posix()
                rec_map = rec_map_col_realpath[_realpath]
                assert rec_map.key_field == "_id" and rec_map.add_filepath_field
                corpus_rec_maps.append(rec_map)

            self.corpus = rows_by_key.RowsByKey(
                record_mapping=corpus_rec_maps,
                allow_duplicates_per_id_lazy=False,
                num_proc=num_proc,
            )

            query_rec_maps = list()
            for path in self.args.query_path:
                _realpath = Path(path).resolve().absolute().resolve().as_posix()
                rec_map = rec_map_col_realpath[_realpath]
                assert rec_map.key_field == "_id" and rec_map.add_filepath_field
                query_rec_maps.append(rec_map)

            self.queries = rows_by_key.RowsByKey(
                record_mapping=query_rec_maps,
                allow_duplicates_per_id_lazy=False,
                num_proc=num_proc,
            )

        else:
            logger.info("Reading files to create MaterializedQRel.corpus")
            # All passage records accessible by docid
            self.corpus = rows_by_key.RowsByKey(
                filepath=self.args.corpus_path,
                new_key_field="_id",
                add_filepath_field=True,
                allow_duplicates_per_id_lazy=False,
                num_proc=num_proc,
            )

            logger.info("Reading files to create MaterializedQRel.queries")
            # All query records accessible by qid
            self.queries = rows_by_key.RowsByKey(
                filepath=self.args.query_path,
                new_key_field="_id",
                add_filepath_field=True,
                allow_duplicates_per_id_lazy=False,
                num_proc=num_proc,
            )
            logger.debug("Finished creating RowsByKey for corpus files.")

        # ID of all queries that we have related docs for
        self.all_qids_local = list(sorted(list(self.qrel_mapping.keys())))
        # If you are working with multiple instances of `MaterializedQRel` that contain conflicting IDs,
        # You should always use global IDs to interact with this class form outside.
        self.all_qids_global = self.get_global_qids()

    @staticmethod
    def create_score_transform(
        args_score_transform: Optional[
            Union[str, int, float, Callable[[Dict[str, Any]], Union[int, float]]]
        ]
    ) -> Callable[[Dict[str, Any]], Union[int, float]]:
        """Create score transform function from class arguments.

        This method should not be used when applying the ``score_transform`` function.
        Instead, you should call this in `__init__` to generate the score_transform callable
        and use that in the rest of the code.

        Args:
            args_score_transform: value of ``score_transform`` in ``MaterializedQRelConfig`` object.

        Returns:
            The score transform function that should be applied to each query/document score.
        """
        if args_score_transform is None:
            score_transform_fn = lambda x: x["score"]
        elif callable(args_score_transform):
            score_transform_fn = args_score_transform
        elif isinstance(args_score_transform, (int, float)):
            score_transform_fn = lambda _: args_score_transform
        elif isinstance(args_score_transform, str):
            if args_score_transform == "floor":
                score_transform_fn = lambda x: int(x["score"])
            elif args_score_transform == "ceil":
                score_transform_fn = lambda x: math.ceil(x["score"])
            else:
                raise ValueError
        else:
            raise TypeError
        return score_transform_fn

    @staticmethod
    def create_group_filter_fn(
        args_group_filter_fn: Optional[
            Callable[[List[Dict[str, Any]]], List[Dict[str, Any]]]
        ],
        group_top_k: Optional[int],
        group_bottom_k: Optional[int],
        group_first_k: Optional[int],
        group_random_k: Optional[int],
    ) -> Optional[Callable[[List[Dict[str, Any]]], List[Dict[str, Any]]]]:
        """Create ``group_filter_fn`` function from class arguments.

        You should call this in `__init__` to generate the ``group_filter_fn`` callable
        and use that in the rest of the code.

        Arguments are a subset of :py:class:`trove.containers.materialized_qrel_config.MaterializedQRelConfig` attributes. See its docstring
        for that class for details.
        """
        group_filter_fn = args_group_filter_fn
        if group_filter_fn is None:
            _non_empty_args = [
                group_top_k is not None,
                group_bottom_k is not None,
                group_first_k is not None,
                group_random_k is not None,
            ]
            assert sum(_non_empty_args) <= 1
            if group_first_k is not None:
                # choose first k documents in their original order
                group_filter_fn = lambda triplets: triplets[:group_first_k]
            elif group_bottom_k is not None:
                # choose k docs with lowest scores
                group_filter_fn = lambda triplets: sorted(
                    triplets, key=lambda x: x["score"]
                )[:group_bottom_k]
            elif group_top_k is not None:
                # choose k docs with largest scores
                group_filter_fn = lambda triplets: sorted(
                    triplets, reverse=True, key=lambda x: x["score"]
                )[:group_top_k]
            elif group_random_k is not None:
                group_filter_fn = lambda recs: (
                    [
                        recs[i]
                        for i in np.random.choice(
                            len(recs), [group_random_k], replace=False
                        )
                    ]
                    if len(recs) > group_random_k
                    else recs
                )
        return group_filter_fn

    def update_metadata(self) -> None:
        """Updates the metadata for the container.

        It creates a new fingerprint, ``cache_dir``, and metadata dict for the container.
        """
        # The combined hash of the following attributes uniquely identifies this container
        # A container with the same hash always produces the same output
        hasher = Hasher()
        for k in sorted(self.file_name_to_id):
            hasher.update(k)
            hasher.update(self.file_name_to_id[k])
        hasher.update(self.qrel_mapping.fingerprint)
        hasher.update(self.score_transform)
        if self.group_filter_fn is not None:
            hasher.update(self.group_filter_fn)
        hasher.update(self.corpus.fingerprint)
        hasher.update(self.queries.fingerprint)
        self._unique_fingerprint = hasher.hexdigest()
        self._info = {
            "fingerprint": self._unique_fingerprint,
            "args": self.args.to_dict(),
            "file_name_to_id": {
                Path(k).as_posix(): v for k, v in self.file_name_to_id.items()
            },
        }
        cache_dir = Path(
            cache_manager.get_cache_dir(
                artifact_content="materialized_qrel_cache",
                artifact_type="final",
                fingerprint=self.fingerprint,
                metadata=self.info,
            )
        )
        cache_dir.mkdir(exist_ok=True, parents=True)
        self._cache_dir = cache_dir.as_posix()

    @property
    def fingerprint(self) -> str:
        """A unique fingerprint for the contents and output of this container.

        Containers with the same fingerprint generate the same output.
        """
        if self._unique_fingerprint is None:
            self.update_metadata()
        return self._unique_fingerprint

    @property
    def info(self) -> Dict:
        if self._info is None or self._info["fingerprint"] != self.fingerprint:
            self.update_metadata()
        return self._info

    @property
    def cache_dir(self) -> Path:
        if self._cache_dir is None or self.info["fingerprint"] != self.fingerprint:
            self.update_metadata()
        return Path(self._cache_dir)

    def set_index_lookup_storage_type(self, storage: str) -> None:
        """Select if the key to row index lookup table should be stored in memory or in memory-
        mapped lmdb dict."""
        self.qrel_mapping.set_index_lookup_storage_type(storage)
        self.queries.set_index_lookup_storage_type(storage)
        self.corpus.set_index_lookup_storage_type(storage)

    def local_to_global_id(self, _id: str, _file: os.PathLike, **_) -> str:
        """Calculate the unique global ID across files for each local ID.

        The global ID is the local ID (i.e., original ID) plus the unique file ID appended at the end:
        ``global_id = original_id + '_' + unique_file_id``

        Args:
            _id (str): local id to generate a global id from
            _file (os.PathLike): the file that contains the record with this ID
            **_: Not used. Just to capture extra arguments

        Returns:
            The global ID (unique across files) corresponding the local ``_id``
        """
        return f"{_id}_{self.file_name_to_id[_file]}"

    def get_global_qids(self) -> List[str]:
        """Created global IDs for all queries."""
        logger.debug("Create global qids")
        with self.queries.meta_info_only():
            global_qids = [
                self.local_to_global_id(_id=qid, _file=self.queries[qid][0]["_file"])
                for qid in self.all_qids_local
            ]
        return global_qids

    def update_file_name_to_id(
        self, name_to_id_mapping: Dict[os.PathLike, str]
    ) -> None:
        """Assign new global IDs to files used in this class.

        Args:
            name_to_id_mapping (Dict[os.PathLike, str]): A mapping from filepath to its global ID.
        """
        # Since different paths might refer to the same file (e.g., symlinks),
        # use the resolved absolute path (similar to linux realpath command) to refer to files
        name_to_id_mapping_realpath = {
            Path(k).resolve().absolute().resolve().as_posix(): v
            for k, v in name_to_id_mapping.items()
        }

        # Find the new file IDs for files used in this class.
        new_mapping = dict()
        for key in self.file_name_to_id.keys():
            realpath_key = Path(key).resolve().absolute().resolve().as_posix()
            new_mapping[key] = name_to_id_mapping_realpath[realpath_key]

        if not all(
            [
                self.file_name_to_id[k] == new_mapping[k]
                for k in self.file_name_to_id.keys()
            ]
        ):
            # if IDs have changed, update the mapping and recalculate the global query IDs
            self.file_name_to_id = new_mapping
            if self.all_qids_global is not None:
                self.all_qids_global = self.get_global_qids()

    def _get_qrel_triplets(
        self, qid: str, default: Optional[Any] = None, strict: bool = True
    ) -> List[Dict[str, Union[str, float]]]:
        """Loads the qrel triplets for a given query.

        It takes care of necessary preprocessings like flattening grouped qrels and applying
        self.score_transform.

        Args:
            qid (str): ID of the query to load its related records.
            default (Optional[Any]): the default value for ``qrel_mapping.get()`` method.
            strict (bool): if true, use ``qrel_mapping[qid]``. Otherwise use ``qrel_mapping.get(qid, default)``

        Returns:
            A list of qrel triplets, where each item is a dict with three keys: ``qid``, ``docid``, and ``score``
        """

        if strict:
            qrel_recs = self.qrel_mapping[qid]
        else:
            qrel_recs = self.qrel_mapping.get(qid, default)
        qrel_recs = container_utils.flatten_grouped_triplets(qrel_recs)
        if self.group_filter_fn is not None:
            qrel_recs = self.group_filter_fn(qrel_recs)
        qrel_recs = [{**rec, "score": self.score_transform(rec)} for rec in qrel_recs]
        return qrel_recs

    def _get_related_recs(
        self,
        qid: str,
        materialize_full: bool = False,
        materialize_meta_info: bool = False,
    ) -> Tuple[Dict, List[Dict]]:
        """Get the query record and related document records for some qid.

        * ``query_rec`` is a dict with an ``_id`` field and if ``materialize_full`` or ``materialize_meta_info`` is ``True``, it also contains other fields
        such as ``text``, ``_file``, etc.

        * ``doc_recs`` is a list of all the documents related to ``qid`` based on the qrel files.
        I.e., list of documents that we have a corresponding (``qid``, ``docid``, ``score``) triplet for them.
        Each document is a dict with ``_id`` and ``score`` fields.If ``materialize_full`` or ``materialize_meta_info``
        is ``True``, it also includes additional fields like ``title``, ``text``, ``file``, etc.

        **The ``_id`` field of these records contain the local record IDs. Should not be used outside of this class.**

        Args:
            qid (str): ID of the query to load its related records.
            materialize_full (bool): If true, load the full content of query/document with
                all available fields (e.g. text) as well.
            materialize_meta_info (bool): If True, load the metadata fields for query/document (e.g., ``_file``)

        Returns:
            A tuple. First item is the corresponding query record and the second item
            is a list of related document records. If both ``materialize_full`` and ``materialize_meta_info``
            are ``False``, it only return the IDs and scores, without the main contentof query/document.
        """
        # corresponding triplets for qid
        qrel_recs = self._get_qrel_triplets(qid=qid, strict=True)
        if not len(qrel_recs):
            msg = f"There are no qrel triplets for query with qid: '{qid}'"
            raise ValueError(msg)

        if materialize_full or materialize_meta_info:
            with ExitStack() as stack:
                if not materialize_full:
                    stack.enter_context(self.queries.meta_info_only())
                    stack.enter_context(self.corpus.meta_info_only())

                # Extract the content of each query and document
                query_rec = self.queries[qid][0]
                doc_recs = list()
                for qrel_rec in qrel_recs:
                    rec = self.corpus[qrel_rec["docid"]][0]
                    doc_recs.append({**rec, "score": qrel_rec["score"]})
        else:
            # Only return _id and score
            query_rec = {"_id": qid}
            doc_recs = [
                {"_id": item["docid"], "score": item["score"]} for item in qrel_recs
            ]
        return query_rec, doc_recs

    def get_related_recs_for_local_qid(
        self,
        qid: str,
        materialize: bool = False,
        return_global_ids: bool = False,
        strict: bool = True,
    ) -> Tuple[Optional[Dict], Optional[List[Dict]]]:
        """Retrieve the related query and document records for a given ``qid``.

        See ``MaterializedQRel._get_related_recs`` for more details.

        Args:
            qid (str): See ``MaterializedQRel._get_related_recs`` for more details.
            materialize (bool): Even if this is false, the records could be materialized if ``return_global_ids`` is true.
                See ``self._get_related_recs()`` for more details.
            return_global_ids (bool): Whether to return global or local IDs in records
            strict (bool): Decides what to do if there are no corresponding qrel triplets for this qid.
                In such cases, if ``strict == False``, it returns a tuple of ``(None, None)``, and if ``strict == True`` , it raises an exception.

        Returns:
            A tuple of query record and related document records. See `self._get_related_recs()` for more details.
        """
        if not strict and qid not in self.all_qids_local:
            # There are no triplets for this query
            return None, None

        # Query and doc recs with local IDs
        query_rec, doc_recs = self._get_related_recs(
            qid=qid,
            materialize_full=materialize,
            materialize_meta_info=return_global_ids,
        )

        if return_global_ids:
            # Calculate and replace local IDs with global IDs
            if not materialize:
                # Create new records without additional fields if not asked for materialized records
                query_rec = {"_id": self.local_to_global_id(**query_rec)}
                for i in range(len(doc_recs)):
                    doc_recs[i] = {
                        "score": doc_recs[i]["score"],
                        "_id": self.local_to_global_id(**doc_recs[i]),
                    }
            else:
                # Only replace the local ids with global and keep other fields.
                query_rec["_id"] = self.local_to_global_id(**query_rec)
                query_rec.pop("_file", None)
                for i in range(len(doc_recs)):
                    doc_recs[i]["_id"] = self.local_to_global_id(**doc_recs[i])
                    doc_recs[i].pop("_file", None)
        elif materialize:
            # Remove the extra field from records
            query_rec.pop("_file", None)
            for i in range(len(doc_recs)):
                doc_recs[i].pop("_file", None)

        return query_rec, doc_recs

    def get_related_recs_for_global_qid(
        self,
        qid: str,
        materialize: bool = False,
        return_global_ids: bool = False,
        strict: bool = True,
    ) -> Tuple[Optional[Dict], Optional[List[Dict]]]:
        """Get related query and document records given the global id a query.

        See ``self.get_related_recs_for_local_qid()`` for details.

        Args:
            strict (bool): Decides what to do if there are no corresponding qrel triplets for this ``qid``.
                In such cases, if ``strict == False``, it returns a tuple of ``(None, None)``, and if ``strict == True``, it raises an exception.

        Returns:
            A tuple of query record and related document records.
            And ``(None, None)`` if there are no triplets for query with the given ID (when ``strict == False``).
            See ``self.get_related_recs_for_local_qid()`` for more details.
        """
        # Make sure to use the **global_qid** to check if we have any corresponding triplets for this query
        # do NOT use local_qid. local IDs across MaterializedQRel instances are not necessarily unique
        # and this method is intended to be called by other classes that work with multiple instances of MaterializedQRel
        if qid not in self.all_qids_global:
            # We do not have any records for this query
            if strict:
                msg = f"There are no qrel triplets for query with global qid '{qid}'"
                raise ValueError(msg)
            else:
                return None, None
        qid = qid.rsplit("_", maxsplit=1)[0]
        return self.get_related_recs_for_local_qid(
            qid=qid,
            materialize=materialize,
            return_global_ids=return_global_ids,
            strict=strict,  # no impact. But, pass the logic along for consistency, making further development easier
        )

    def get_related_recs(
        self,
        global_qid: Optional[str] = None,
        local_qid: Optional[str] = None,
        materialize: bool = False,
        return_global_ids: bool = True,
        strict: bool = True,
    ) -> Tuple[Optional[Dict], Optional[List[Dict]]]:
        """Get the related records for some query.

        If ``strict==False`` and there are no triplets for query with the given ID, it returns ``(None, None)``.
        See ``MaterializedQRel.get_related_recs_for_local_qid()`` for more details.

        Args:
            global_qid (Optional[str]): global ID of the query. Takes precedence over ``local_qid`` if provided.
            local_qid (Optional[str]): local ID of the query. It is ignored if ``global_qid`` is provided.
        """
        if (global_qid is None) == (local_qid is None):
            msg = "You should provide exactly one of the 'global_qid' or 'local_qid' arguments."
            raise ValueError(msg)

        if global_qid is not None:
            return self.get_related_recs_for_global_qid(
                qid=global_qid,
                materialize=materialize,
                return_global_ids=return_global_ids,
                strict=strict,
            )
        else:
            return self.get_related_recs_for_local_qid(
                qid=local_qid,
                materialize=materialize,
                return_global_ids=return_global_ids,
                strict=strict,
            )

    def get_qrel_nested_dict(
        self, return_global_ids: bool = False
    ) -> Dict[str, Dict[str, Union[int, float]]]:
        """Converts the qrel triplets to the nested dict format used by ``pytrec_eval``.

        Args:
            return_global_ids (bool): if true, use global ids for queries and documents.

        Returns:
            a nested dict where dict[qid][docid] is the score between query ``qid`` and document ``docid``.
        """
        logger.debug("Create nested qrel dict for MaterializedQRel.")
        # choose the correct cache file
        if return_global_ids:
            cache_file = self.cache_dir / "nested_qrel_with_global_ids.pkl"
        else:
            cache_file = self.cache_dir / "nested_qrel_with_local_ids.pkl"

        qrel_mapping = None
        with file_utils.easyfilelock(cache_file.as_posix()):
            if not cache_file.exists():
                qrel_mapping = defaultdict(dict)
                with self.queries.meta_info_only(), self.corpus.meta_info_only():
                    for qid in tqdm(
                        self.all_qids_local,
                        desc="Make Qrel Dict",
                        disable=not logging_conf.is_debug(),
                    ):
                        rel_docs = self._get_qrel_triplets(qid=qid, strict=True)
                        if return_global_ids:
                            rec_qid = self.local_to_global_id(**self.queries[qid][0])
                        else:
                            rec_qid = qid
                        for doc in rel_docs:
                            if return_global_ids:
                                docid = self.local_to_global_id(
                                    **self.corpus[doc["docid"]][0]
                                )
                            else:
                                docid = doc["docid"]
                            qrel_mapping[rec_qid][docid] = doc["score"]
                qrel_mapping = dict(qrel_mapping)
                with file_utils.atomic_write(file=cache_file, root="parent") as tfile:
                    with open(tfile, "wb") as f:
                        pkl.dump(qrel_mapping, f)

        if qrel_mapping is None:
            with open(cache_file, "rb") as f:
                qrel_mapping = pkl.load(f)
        return qrel_mapping



#!/usr/bin/env python

# from here:
# https://github.com/huggingface/datasets/blob/a1b5a32365cd2c6631fa4ad236a881f9f62bf32b/src/datasets/commands/datasets_cli.py
from argparse import ArgumentParser

from trove.commands.env import EnvironmentCommand


def parse_unknown_args(unknown_args):
    return {
        key.lstrip("-"): value
        for key, value in zip(unknown_args[::2], unknown_args[1::2])
    }


def main():
    parser = ArgumentParser(
        "Trove CLI tool", usage="trove-cli <command> [<args>]", allow_abbrev=False
    )
    commands_parser = parser.add_subparsers(help="trove-cli command helpers")

    # Register commands
    EnvironmentCommand.register_subcommand(commands_parser)

    # Parse args
    args, unknown_args = parser.parse_known_args()
    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)
    kwargs = parse_unknown_args(unknown_args)

    # Run
    service = args.func(args, **kwargs)
    service.run()


if __name__ == "__main__":
    main()



import itertools
import os
import pickle as pkl
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import datasets
import pyarrow as pa
from datasets.formatting import query_table
from tqdm import tqdm

from .. import cache_manager, file_utils
from ..data import polars_transforms
from ..logging_utils import get_logger_with_config

logger, logging_conf = get_logger_with_config("trove")


def dataset_column_to_row_indices_mapping_op(
    dataset: datasets.Dataset,
    id_field: str,
    dst_file: os.PathLike,
    num_proc: Optional[int] = None,
) -> None:
    """Creates a mapping from values of a column (i.e., keys) to row indices for the given dataset.

    You should not use this function directly. Instead, use ``dataset_column_to_row_indices_mapping``.

    Writes the results to a pickle file that contains a dict
    from from values of ``id_field`` to a list of indices of rows that contain that value.

    We save them as a pickle file to avoid the cost of creating a ``dict`` object when loading it.

    Args:
        dataset (datasets.Dataset): dataset to create the index mapping for.
        id_field (str): name of the field to use as key.
        dst_file (os.PathLike): path to file where the resulting pickle file will be saved.
        num_proc (Optional[int]): number of parallel processes to use. Passed to ``datasets.Dataset.*`` methods.
    """
    dst_file = Path(dst_file)
    if dst_file.suffix not in [".pkl", ".pickle"]:
        msg = f"We only support writing to pickle file. But, got: {dst_file.as_posix()}"
        raise ValueError(msg)

    with file_utils.easyfilelock(dst_file):
        if not Path(dst_file).exists():
            keys_cache = cache_manager.get_cache_dir(
                input_data=dataset,
                artifact_content="key_to_row_idx_polars",
                artifact_type="intermediate",
                fingerprint_src="path",
            )
            # we do not need to lock this file for writing because it is only written to
            # in this function and while the lock for dst_cache is acquired.
            keys_cache = Path(keys_cache, "dataset_rows_with_key_only.arrow").as_posix()
            Path(keys_cache).parent.mkdir(exist_ok=True, parents=True)
            # we only need the keys. Ignore other columns
            key_ds = dataset.select_columns(id_field)

            batch_size = 1_000
            with pa.OSFile(keys_cache, "wb") as sink:
                with pa.ipc.new_file(sink, key_ds.features.arrow_schema) as writer:
                    # write in batches. It is more memory efficient
                    # adapted from here: https://github.com/huggingface/datasets/blob/01f91bae037c98f2e05456287bab21470adb8f07/src/datasets/arrow_dataset.py#L4951
                    for offset in tqdm(
                        range(0, len(key_ds), batch_size),
                        desc="Write dataset keys in arrow table",
                        disable=not logging_conf.is_debug(),
                    ):
                        sub_tab = query_table(
                            table=key_ds._data,
                            key=slice(offset, offset + batch_size),
                            indices=(
                                key_ds._indices if key_ds._indices is not None else None
                            ),
                        )
                        writer.write_table(sub_tab, 1_000)
            # It is recommended to use atomic_write right before starting the write operation
            # In this special case, it is okay to call it earlier because we know 'polars_transforms.key_to_row_indices_mapping_op'
            # does not do any processing/locking based on filepath and just writes to it.
            with file_utils.atomic_write(file=dst_file, root="parent") as tfile:
                polars_transforms.key_to_row_indices_mapping_op(
                    src_file=keys_cache, dst_file=tfile, id_field=id_field
                )


def dataset_column_to_row_indices_mapping(
    dataset: datasets.Dataset,
    id_field: str,
    num_proc: Optional[int] = None,
    load: bool = True,
) -> Union[Dict[Any, List[int]], os.PathLike]:
    """Creates a mapping from values of a column (i.e., keys) to row indices for the given dataset.

    It internally calls ``dataset_column_to_row_indices_mapping_op`` and does the same operation.
    However, this function saves the results to a unique cache file and reads from the cache file if available.
    It optionally returns the loaded mapping as a dictionary.

    Args:
        dataset (datasets.Dataset): dataset to create index mapping for.
        id_field (str): name of the field/column to use as key.
        num_proc (Optional[int]): number of parallel processes used. Arg passed to ``datasets.Dataset.*`` methods.
        load (bool): Whether to return the mapping object or return the path to the pickle file holding it.

    Returns:
        A mapping from the values of ``id_field`` column to indices of rows that contain that specific value.
        If ``load`` is ``False``, it returns the path to a pickle file containing the mapping instead of loading it.
    """
    logger.debug(
        f"Create key to row indices mapping. num records: {tqdm.format_sizeof(len(dataset))}"
    )

    cache_file = cache_manager.get_cache_dir(
        input_data=dataset,
        artifact_type="final",
        artifact_content="key_to_row_indices_mapping",
    )
    cache_file = Path(cache_file, "key2row_idx.pkl").as_posix()
    dataset_column_to_row_indices_mapping_op(
        dataset=dataset, id_field=id_field, dst_file=cache_file, num_proc=num_proc
    )
    if load:
        with open(cache_file, "rb") as f:
            key_to_indices = pkl.load(f)
    else:
        key_to_indices = cache_file

    return key_to_indices


def flatten_grouped_triplets(
    grouped_triplets: Dict[str, Union[List, Any]]
) -> List[Dict]:
    """Converts a dict of column values to a list of row records.

    You can use this to flatten the grouped qrels for each query.

    In ``grouped_triplets`` dictionary, all values of type ``list`` should be of the same length.
    Values that are not of type `list` are repeated for all rows.

    Example Input:

    .. code-block:: text

        {'a': [1, 2], 'b': ['x', 'y'], 'c': 'foo'}

    Example Output:

    .. code-block:: text

        [
            {'a': 1, 'b': 'x', 'c': 'foo'},
            {'a': 2, 'b': 'y', 'c': 'foo'}
        ]

    Args:
        grouped_triplets (Dict[str, Optional[Union[str, float, List]]]): dict from
            column names to column values (either a list of values or a constant values).

    Returns:
        a list row records.
    """
    qrel_recs = list()
    for grouped_recs in grouped_triplets:
        # Repeat values that are the same for all triplets to get a dict of lists
        _grecs = {
            k: v if isinstance(v, list) else itertools.repeat(v)
            for k, v in grouped_recs.items()
        }
        # Turn the dict of lists to list of dicts
        qrel_recs.extend(
            [
                dict(zip(ks, vs))
                for ks, vs in zip(
                    itertools.repeat(tuple(_grecs.keys())), zip(*_grecs.values())
                )
            ]
        )
    return qrel_recs


def filter_fn_wrapper(fn: Callable[[Dict[str, Union[str, float]]], bool]) -> Callable:
    """Wrapper that applies a given function that filters individual triplets over grouped
    triplets.

    You can use this function to apply ``fn`` to grouped triplets using ``datasets.Dataset.map``.

    Args:
        fn (Callable[[Dict[str, Union[str, float]]], bool]): callable that takes a dict
            of form ``{'qid': ..., 'docid': ..., 'score': ...}`` and returns a boolean.
            If returned true, we keep the triplet. And skip it otherwise.

    Returns:
        A function that applies ``fn`` over grouped triplets.
    """

    def filter_groups(groups: Dict[str, List]) -> Dict[str, List]:
        """Filtering a batch of grouped triplets."""
        groups_subset = {"qid": [], "docid": [], "score": []}
        for g_qid, g_docids, g_scores in zip(
            groups["qid"], groups["docid"], groups["score"]
        ):
            subset_docids = list()
            subset_scores = list()
            for _d, _s in zip(g_docids, g_scores):
                if fn(dict(qid=g_qid, docid=_d, score=_s)):
                    subset_docids.append(_d)
                    subset_scores.append(_s)
            # Only keep this group if it contains at least one triplet
            if len(subset_docids):
                groups_subset["qid"].append(g_qid)
                groups_subset["docid"].append(subset_docids)
                groups_subset["score"].append(subset_scores)
        return groups_subset

    return filter_groups



"""Abstraction on top of dense encoders for information retrieval."""

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Optional

import torch
import torch.distributed as dist
from torch import nn
from transformers import TrainingArguments
from transformers.file_utils import ModelOutput

from . import modeling_utils
from .loss_base import RetrievalLoss
from .model_args import ModelArguments
from .pretrained_encoder import PretrainedEncoder


@dataclass
class RetrieverOutput(ModelOutput):
    """Contains the output of the retriever model."""

    query: Optional[torch.Tensor] = None
    """Query embeddings"""
    passage: Optional[torch.Tensor] = None
    """Passage embeddings"""
    loss: Optional[torch.Tensor] = None
    """Calculated loss"""
    logits: Optional[torch.Tensor] = None
    """similarity score between queries and passages"""


class PretrainedRetriever(nn.Module):
    def __init__(
        self,
        model_args: ModelArguments,
        encoder: nn.Module,
        preprocess_only: bool = False,
        format_query: Optional[Callable] = None,
        format_passage: Optional[Callable] = None,
        append_eos_token: Optional[bool] = None,
        loss_extra_kwargs: Optional[Dict] = None,
    ) -> None:
        """A base class for training/inference with different retrievers.

        Args:
            model_args (ModelArguments): config specifying the model and loss to use.
                Currently, we only use ``model_args`` to instantiate the loss (it might also be saved by some loggers).
                To instantiate both model and loss, use :meth:`from_model_args`.
            encoder (nn.Module): encoder model to use. It must expose ``encode_query()`` and ``encode_passage()`` methods.
                ``encoder`` is also expected to provide a ``save_pretrained()`` method but only if you call :meth:`save_pretrained`
            preprocess_only (bool): if true, do not instantiate loss module.
                You should also pass this to :meth:`from_model_args` if you
                do not want to load the model parameters. See ``PretrainedEncoder.__init__`` for details.
            format_query (Optional[Callable]): Callable similar to `PretrainedEncoder.format_query`.
                If provided, it is prioritized over ``encoder.format_query()``.
                It is not used by this class internally. It is just exposed as a convenience method to
                keep everything needed to encode a query in one place.
            format_passage (Optional[Callable]): Callable similar to ``PretrainedEncoder.format_passage()``
                If provided, it is prioritized over ``encoder.format_passage``.
                It is not used by this class internally. It is just exposed as a convenience method to
                keep everything needed to encode a query in one place.
            append_eos_token (Optional[bool]): Similar to ``PretrainedEncoder.append_eos_token``
                If provided, it is prioritized over ``encoder.append_eos_token``.
                It is not used by this class internally. It is just exposed as a convenience method to
                keep everything needed to encode a query in one place.
            loss_extra_kwargs (Optional[Dict]): If given, these are passed to ``RetrievalLoss.__init__()`` as keyword arguments.
        """
        super().__init__()

        self.args = model_args
        self.encoder = encoder

        # The encoder model should take care of these
        # Expose these to remain compatible with a subset (and *NOT* all) of tools
        # (e.g., transformers.Trainer) that expect a transformers.PretrainedModel instance
        modeling_utils.add_model_apis_to_wrapper(wrapper=self, model=encoder)

        if not preprocess_only:
            _model_param = next(iter(self.encoder.parameters()))
            _loss_kwargs = dict(dtype=_model_param.dtype, device=_model_param.device)
            if loss_extra_kwargs is not None:
                _loss_kwargs = {**_loss_kwargs, **loss_extra_kwargs}
            self.loss: RetrievalLoss = RetrievalLoss.from_model_args(
                args=self.args, **_loss_kwargs
            )
        else:
            self.loss = None

        if format_query is not None:
            self.format_query = format_query
        else:
            self.format_query = getattr(encoder, "format_query", None)

        if format_passage is not None:
            self.format_passage = format_passage
        else:
            self.format_passage = getattr(encoder, "format_passage", None)

        if append_eos_token is not None:
            self.append_eos_token = append_eos_token
        else:
            self.append_eos_token = getattr(encoder, "append_eos_token", None)

        if hasattr(encoder, "similarity_fn"):
            self.similarity_fn = encoder.similarity_fn
        else:
            self.similarity_fn = lambda x, y: x @ y.T

        self.process_rank = None
        self.world_size = None
        self.is_ddp = dist.is_initialized()
        if self.is_ddp:
            self.process_rank = dist.get_rank()
            self.world_size = dist.get_world_size()

    @classmethod
    def from_model_args(
        cls,
        args: ModelArguments,
        model_name_or_path: Optional[str] = None,
        training_args: Optional[TrainingArguments] = None,
        loss_extra_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        """Instantiate the retriever according based on the given args.

        Args:
            args (ModelArguments): config used to instantiate the encoder and loss modules.
            model_name_or_path (Optional[str]): name of the encoder model to load.
                If not provided, use ``args.model_name_or_path``. You should almost never
                need to use this.
            training_args (TrainingArguments): passed to ``PretrainedEncoder.from_model_args()``.
                It is used to enable gradient checkpointing if needed.
            loss_extra_kwargs (Optional[Dict]): passed to ``BiEncoderRetriever.__init__()`` which then passes them to ``RetrievalLoss.__init__()``.
            **kwargs: extra keyword arguments passed to ``BiEncoderRetriever.__init__()`` and
                ``PretrainedEncoder.from_model_args()``. If you want to avoid loading model parameters
                and only load methods and attributes required for pre-processing, you should pass
                ``preprocess_only=True`` as part of this kwargs. See ``BiEncoderRetriever.__init__()`` and ``PretrainedEncoder.__init__()``
                for details.
        Returns:
            an instance of one of `PretrainedRetriever` subclasses.
        """
        if args.encoder_class is None or args.encoder_class.lower() == "none":
            m_name = (
                args.model_name_or_path
                if model_name_or_path is None
                else model_name_or_path
            )
            trove_conf = modeling_utils.load_trove_retriever_config(
                model_name=m_name, model_revision=args.model_revision
            )
            if trove_conf is not None and "encoder_class" in trove_conf:
                args.encoder_class = trove_conf["encoder_class"]

        # separate kwargs that are only used for BiEncoderRetriever.__init__()
        cls_kwargs = dict()
        for k in ["format_query", "format_passage", "append_eos_token"]:
            if k in kwargs:
                cls_kwargs[k] = kwargs.pop(k)
        # this one is used by both retriever and encoder init methods.
        # So, do not '.pop' it from kwargs
        if "preprocess_only" in kwargs:
            cls_kwargs["preprocess_only"] = kwargs["preprocess_only"]

        encoder: PretrainedEncoder = PretrainedEncoder.from_model_args(
            args=args,
            model_name_or_path=model_name_or_path,
            training_args=training_args,
            **kwargs,
        )

        retriever = cls(
            model_args=args,
            encoder=encoder,
            loss_extra_kwargs=loss_extra_kwargs,
            **cls_kwargs,
        )
        return retriever

    def forward(
        self,
        query: Optional[Dict[str, torch.Tensor]] = None,
        passage: Optional[Dict[str, torch.Tensor]] = None,
        label: Optional[torch.Tensor] = None,
        return_loss: Optional[bool] = True,
        **kwargs,
    ) -> RetrieverOutput:
        """Encodes query and passages and potentially calculate similarity scores and loss.

        Args:
            query (Optional[Dict[str, torch.Tensor]]): tokenized query.
            passage (Optional[Dict[str, torch.Tensor]]): tokenized passages.
            label (Optional[torch.Tensor]): Relevancy scores of the corresponding passages for each query.
                If there are k documents for each query, this is a 2D tensor of shape `[NUM_QUERIES, k]`
            return_loss (Optional[bool]): if true, calculate the loss value.
            **kwargs: unused keyword arguments are passed to the ``forward()`` method of the loss module.

        Returns:
            If possible, query and passage embeddings as well as similarity and loss scores.
        """
        raise NotImplementedError

    def save_pretrained(self, *args, **kwargs):
        is_main_process = True
        if "is_main_process" in kwargs:
            is_main_process = kwargs["is_main_process"]
        elif len(args) > 1:
            is_main_process = args[1]
        if is_main_process:
            if "save_directory" in kwargs:
                pardir = Path(kwargs["save_directory"])
            else:
                pardir = Path(args[0])
            pardir.mkdir(exist_ok=True, parents=True)
            conf = {"encoder_class": self.encoder.__class__.__name__}
            with open(pardir / modeling_utils.TROVE_RETRIEVER_CONFIG_FILE, "w") as f:
                json.dump(conf, f, indent=2)

        if "state_dict" in kwargs:
            encoder_prefix = "encoder."
            loss_prefix = "loss."
            assert all(
                k.startswith(encoder_prefix) or k.startswith(loss_prefix)
                for k in kwargs["state_dict"].keys()
            )
            kwargs["state_dict"] = {
                k[len(encoder_prefix) :]: v
                for k, v in kwargs["state_dict"].items()
                if not k.startswith(loss_prefix)
            }
        return self.encoder.save_pretrained(*args, **kwargs)



"""Loader functions to read various file formats."""

import json
import os
import pickle as pkl
import re
from pathlib import Path
from typing import List, Optional

import datasets

from .. import cache_manager, file_utils
from ..data import file_transforms
from ..logging_utils import get_logger_with_config
from . import file_reader

logger, logging_conf = get_logger_with_config("trove")


@file_reader.register_loader("qrel")
def qrel_from_grouped_triplets(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Load grouped qrel triplets from JSONL file.

    Each line should have three fields::

        {
            'qid': '...',
            'docid': ['docid1', 'docid2', ...],
            'score': [score1, score2, ...]
        }

    Args:
        filepath (os.PathLike): Path to a JSONL file with grouped qrel triplets
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        huggingface ``datasets.Dataset`` with ``'qid'``, ``'docid'``, and ``'score'`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    # check the given file contains grouped qrels in jsonl format
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix != ".jsonl":
        return
    with open(filepath, "r") as f:
        rec = json.loads(f.readline())
    if not {"qid", "docid", "score"}.issubset(set(list(rec.keys()))):
        return
    if not (
        isinstance(rec["qid"], str)
        and isinstance(rec["docid"], list)
        and isinstance(rec["score"], list)
    ):
        return

    logger.debug("Read qrel from grouped triplets jsonl file")
    ds = datasets.load_dataset(
        "json", data_files=filepath, split="train", num_proc=num_proc
    )
    return ds


@file_reader.register_loader("qrel")
def qrel_from_csv(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Load qrel file from a delimiter separated file.

    It only supports a file with exactly three columns: ``qid``, ``docid``, ``score``
    If the headers are missing, it is assumed the columns are in the following order: ``['qid', 'docid', 'score']``.

    Args:
        filepath (os.PathLike): path to CSV/TSV file
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        huggingface ``datasets.Dataset`` with ``qid``, ``docid``, and ``score`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    # check if it is a csv or tsv file
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix not in [".csv", ".tsv"]:
        return
    logger.debug("Read qrel from csv")

    # Inspect first five lines to figure out the delimiter
    lines = list()
    with open(filepath, "r") as f:
        for i, line in enumerate(f):
            lines.append(line)
            if i > 10:
                break
    # Check for common delimiters
    if all("\t" in line.strip() for line in lines):
        sep = "\t"
    elif all("," in line.strip() for line in lines):
        sep = ","
    elif all(" " in line.strip() for line in lines):
        sep = " "
    else:
        msg = "Could not identify delimiter. It is not any of the supported ',' '\t' 'SPACE' characters."
        raise RuntimeError(msg)

    if any(len(line.strip().split(sep)) != 3 for line in lines):
        msg = (
            "Only csv/tsv files with exactly three columns ('qid', 'docid', 'score') are supported."
            f" number of columns in file: {len(lines[0].strip().split(sep))}"
        )
        raise RuntimeError(msg)

    cache_file = cache_manager.get_cache_dir(
        input_data=filepath, artifact_type="final", artifact_content="grouped_qrels"
    )
    cache_file = Path(cache_file, "qrel.jsonl").as_posix()

    file_transforms.group_qrel_triplets_from_csv(
        src_file=filepath, dst_file=cache_file, num_proc=num_proc
    )
    ds = datasets.load_dataset("json", data_files=cache_file, split="train")
    return ds


@file_reader.register_loader("qrel")
def qrel_from_pickle(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Load qrel triplets from pickle files.

    The file is expected to contain a single object of type dict.
    ``object[qid][docid]`` is the corresponding score for query ``qid`` and document ``docid``.
    ``qid`` and ``docid`` should be of type `str`.


    Args:
        filepath (os.PathLike): pickle file to load.
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        huggingface datasets.Dataset with ``qid``, ``docid``, and ``score`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix not in [".pkl", ".pickle"]:
        return

    logger.debug("Read qrel from pickle.")

    # Cache results to both reduce memory usage by using datasets arrow tables
    # And also avoid duplicate computation next time
    cache_file = cache_manager.get_cache_dir(
        input_data=filepath, artifact_type="final", artifact_content="grouped_qrels"
    )
    cache_file = Path(cache_file, "qrel.jsonl").as_posix()

    with file_utils.easyfilelock(cache_file):
        if not Path(cache_file).exists():
            with open(filepath, "rb") as f:
                dump = pkl.load(f)

            # Check the data structure and dtypes
            if (
                not isinstance(dump, dict)
                or not isinstance(next(iter(dump.keys())), str)
                or not isinstance(next(iter(dump.values())), dict)
                or not isinstance(next(iter(next(iter(dump.values())).keys())), str)
                or not isinstance(
                    next(iter(next(iter(dump.values())).values())), (int, float)
                )
            ):
                msg = (
                    "Data in pickle file is not correctly formatted. It is expected to be a nested dictionary, where dump[qid][docid]=score(qid, docid)."
                    "both qid and docid are expected to be of type 'str' and their similarity should be of type 'int' or 'float'"
                )
                raise RuntimeError(msg)

            with file_utils.atomic_write(file=cache_file, root="parent") as tfile:
                with open(tfile, "w") as f:
                    for i, (qid, qdata) in enumerate(dump.items()):
                        docids, scores = list(zip(*qdata.items()))
                        row = dict(qid=qid, docid=list(docids), score=list(scores))
                        f.write(json.dumps(row) + "\n")
                        if not i % 1000:
                            f.flush()

    ds = datasets.load_dataset(
        "json", data_files=cache_file, split="train", num_proc=num_proc
    )
    return ds


@file_reader.register_loader("qrel")
def qrel_from_tevatron_training_data(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Convert a tevatron training file to qrel triplets.

    The data is expected to be a JSONL file with the structure that `tevatron <https://github.com/texttron/tevatron>`_ uses for training files.
    Each record (i.e., line) should be::

        {
            'query_id': 'target query id',
            'query': 'text of the query',
            'positive_passages': [{'docid': 'od of pos doc', 'title': 'title of pos doc', 'text': 'text of pos doc'}, ..., ...],
            'negative_passages': [{'docid': 'od of neg doc', 'title': 'title of neg doc', 'text': 'text of neg doc'}, ..., ...]
        }

    When creating (qid, docid, score) triplets, we give a score of ``0`` to all negative passages and a score of ``1`` to all positive passages.

    Args:
        filepath (os.PathLike): path to JSONL file in tevatron training format.
        num_proc (Optional[int]): arg to ``datasets.Dataset.map``

    Returns:
        huggingface datasets.Dataset with ``qid``, ``docid``, and ``score`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    # check if it is tevatron training file
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix != ".jsonl":
        return
    with open(filepath, "r") as f:
        rec = json.loads(f.readline())
    if set(list(rec.keys())) != {
        "query_id",
        "query",
        "positive_passages",
        "negative_passages",
    }:
        return
    if not (
        isinstance(rec["query_id"], str)
        and isinstance(rec["query"], str)
        and isinstance(rec["positive_passages"], list)
        and isinstance(rec["negative_passages"], list)
    ):
        return

    def process_row(row):
        new_row = dict()
        new_row["qid2"] = row["query_id"]
        new_row["docid2"] = list()
        new_row["score2"] = list()

        new_row["docid2"].extend([n["docid"] for n in row["negative_passages"]])
        new_row["score2"].extend([0] * len(row["negative_passages"]))

        new_row["docid2"].extend([p["docid"] for p in row["positive_passages"]])
        new_row["score2"].extend([1] * len(row["positive_passages"]))
        return new_row

    logger.debug("Read tevatron training file and convert to grouped qrel format.")

    ds = datasets.load_dataset("json", data_files=filepath, split="train")
    ds = ds.map(process_row, remove_columns=ds.column_names, num_proc=num_proc)
    ds = ds.rename_columns({"qid2": "qid", "docid2": "docid", "score2": "score"})
    return ds


@file_reader.register_loader("qrel")
def qrel_from_sydir_corpus(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Infer qrel triplets from sydir docids.

    sydir docids are formatted as ``f"{qid}_l_{level}_d_{doc_idx}"``.
    We parse this and use ``level`` as the ``score`` field in the qrel triplets.

    Args:
        filepath (os.PathLike): sydir corpus file
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        huggingface datasets.Dataset with ``qid``, ``docid``, and ``score`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    # check if it is a sydir corpus file
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix != ".jsonl":
        return
    with open(filepath, "r") as f:
        rec = json.loads(f.readline())
    if set(list(rec.keys())) != {"_id", "title", "text"}:
        return
    if "_id" in rec:
        rec_id = rec["_id"]
    elif "docid" in rec:
        rec_id = rec["docid"]
    if not isinstance(rec_id, str):
        msg = "All record IDs should be of type 'str'. Got: {type(rec_id)}"
        raise TypeError(msg)
    # Check if the ID follows sydir docid pattern.
    matches = re.findall(r".*?_l_([0-9].*)_d_([0-9].*)", rec_id)
    # There is only one such match in sydir docid format
    if len(matches) != 1:
        return
    try:
        # If it is a sydir docid, the extracted pieces should be both integers.
        _ = int(matches[0][0])
        _ = int(matches[0][1])
    except:
        return

    logger.debug("Read qrel from synthetic sydir corpus file.")
    cache_file = cache_manager.get_cache_dir(
        input_data=filepath, artifact_type="final", artifact_content="grouped_qrels"
    )
    cache_file = Path(cache_file, "qrel.jsonl").as_posix()

    file_transforms.group_qrel_triplets_from_sydir_corpus(
        src_file=filepath, dst_file=cache_file, num_proc=num_proc
    )
    ds = datasets.load_dataset("json", data_files=cache_file, split="train")
    return ds


@file_reader.register_loader("qid")
def qids_from_queries_jsonl(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> List[str]:
    """Load qids from the original `queries.jsonl` files.

    It expect an ``_id`` field in each record of the JSONL file.

    Args:
        filepath (os.PathLike): ``queries.jsonl`` file to read.
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        A list of query IDs
    """
    # check that the given file contains original query records
    if not Path(filepath).exists():
        return
    if Path(filepath).suffix != ".jsonl":
        return
    with open(filepath, "r") as f:
        rec = json.loads(f.readline())
    if set(list(rec.keys())) != {"_id", "text"}:
        # it has other fields that are not usually found in queries.jsonl files
        # or it is missing the two fields that are usually found in queries.jsonl files.
        return

    logger.debug("Read list of qids from original queries.jsonl format.")

    ds = datasets.load_dataset("json", data_files=filepath, split="train")

    # Make sure the IDs are of type 'str'
    new_features = ds.features.copy()
    if new_features["_id"] != datasets.Value("string"):
        logger.debug("Cast '_id' field to 'str' dtype.")

        new_features["_id"] = datasets.Value("string")
        ds = ds.cast(new_features, num_proc=num_proc)
    return ds["_id"]
